#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        raw_classes.py
# Purpose:     Implements helper classes that contain
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     19-06-2022
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
"""
Defines base classes for the RAW file data structures.
"""
import numpy as np
from numpy import zeros, complex128, float32, float64
from typing import Union, List


class DataSet(object):
    """
    This is the base class for storing all traces of a RAW file. Returned by the get_trace() or by the get_axis()
    methods.
    Normally the user doesn't have to be aware of this class. It is only used internally to encapsulate the different
    implementations of the wave population.
    Data can be retrieved directly by using the [] operator.
    If numpy is available, the numpy vector can be retrieved by using the get_wave() method.
    The parameter whattype defines what is the trace representing in the simulation, Voltage, Current a Time or
    Frequency.
    """

    def __init__(self, name, whattype, datalen, numerical_type='real'):
        """Base Class for both Axis and Trace Classes.
        Defines the common operations between both."""
        self.name = name
        self.whattype = whattype
        self.numerical_type = numerical_type
        if numerical_type == 'double':
            self.data = zeros(datalen, dtype=float64)
        elif numerical_type == 'real':
            self.data = zeros(datalen, dtype=float32)
        elif numerical_type == 'complex':
            self.data = zeros(datalen, dtype=complex128)
        else:
            raise NotImplementedError

    def __str__(self):
        return f"name:'{self.name}'\ntype:'{self.whattype}'\nlen:{len(self.data)}"

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)

    def __getitem__(self, item):
        return self.data[item]

    def get_wave(self) -> np.array:
        """
        :return: Internal data array
        :rtype: numpy.array
        """
        return self.data


class Axis(DataSet):
    """This class is used to represent the horizontal axis like on a Transient or DC Sweep Simulation. It derives from
    the DataSet and defines additional methods that are specific for X axis.
    This class is constructed by the get_time_axis() method or by a get_trace(0) command. In RAW files the trace 0 is
    always the X Axis. Ex: time for .TRAN simulations and frequency for the .AC simulations.

    To access data inside this class, the get_wave() should be used, which implements the support for the STEPed data.
    IF Numpy is available, get_wave() will return a numpy array.

    In Transient Analysis and in DC transfer characteristic, LTSpice uses doubles to store the axis values. QSpice
    uses doubles for all variables.
    """

    def __init__(self, name: str, whattype: str, datalen: int, numerical_type: str = 'double'):
        super().__init__(name, whattype, datalen, numerical_type)
        self.step_info = None

    def _set_steps(self, step_info: List[dict]):
        self.step_info = step_info

        self.step_offsets = [None for _ in range(len(step_info))]

        # Now going to calculate the point offset for each step
        self.step_offsets[0] = 0
        i = 1
        k = 1
        while i < len(self.data):
            if self.data[i] == self.data[0]:
                self.step_offsets[k] = i
                k += 1
            i += 1

        if k != len(self.step_info):
            raise SpiceReadException("The file a different number of steps than expected.\n" +
                                     "Expecting %d got %d" % (len(self.step_offsets), k))

    def step_offset(self, step: int):
        """
        In Stepped RAW files, several simulations runs are stored in the same RAW file. This function returns the
        offset within the binary stream where each step starts.

        :param step: Number of the step within the RAW file
        :type step: int
        :return: The offset within the RAW file
        :rtype: int
        """
        if self.step_info is None:
            if step > 0:
                return len(self.data)
            else:
                return 0
        else:
            if step >= len(self.step_offsets):
                return len(self.data)
            else:
                return self.step_offsets[step]

    def get_wave(self, step: int = 0) -> np.array:
        """
        Returns a vector containing the wave values. If numpy is installed, data is returned as a numpy array.
        If not, the wave is returned as a list of floats.

        If stepped data is present in the array, the user should specify which step is to be returned. Failing to do so,
        will return all available steps concatenated together.

        :param step: Optional step in stepped data raw files.
        :type step: int
        :return: The trace values
        :rtype: numpy.array
        """
        if step == 0:
            wave = self.data[:self.step_offset(1)]
        else:
            wave = self.data[self.step_offset(step):self.step_offset(step + 1)]
        if self.name == 'time':  # This is a bug in LTSpice, where the time axis values are sometimes negative
            return np.abs(wave)
        else:
            return wave

    def get_time_axis(self, step: int = 0):
        """
        .. deprecated:: 1.0 Use `get_wave()` instead.

        Returns the time axis raw data. Please note that the time axis may not have a constant time step. LTSpice will
        increase the time-step in simulation phases where there aren't value changes, and decrease time step in
        the parts where more time accuracy is needed.

        :param step: Optional step number if reading a raw file with stepped data.
        :type step: int
        :return: time axis
        :rtype: numpy.array
        """
        assert self.name == 'time', \
            "This function is only applicable to transient analysis, where a bug exists on the time signal"
        return self.get_wave(step)

    def get_point(self, n: int, step: int = 0) -> Union[float, complex]:
        """
        Get a point from the dataset
        
        :param n: position on the vector
        :type n: int
        :param step: step index, defaults to 0
        :type step: int, optional
        :returns: Value of the data point
        :rtype: float, complex
        """
        return self.data[n + self.step_offset(step)]

    def __getitem__(self, item) -> Union[float, complex]:
        """This is only here for compatibility with previous code. """
        assert self.step_info is None, "Indexing should not be used with stepped data. Use get_point or get_wave"
        return self.data.__getitem__(item)

    def get_position(self, t, step: int = 0) -> Union[int, float]:
        """
        Returns the position of a point in the axis. If the point doesn't exist, an interpolation is done between the
        two closest points.
        For example, if the point requested is 1.0001ms and the closest points that exist in the axis are t[100]=1ms and
        t[101]=1.001ms, then the return value will be 100 + (1.0001ms-1ms)/(1.001ms-1ms) = 100.1

        :param t: point in axis to search for
        :type t: float
        :param step: step number
        :type step: int
        :returns: The position of parameter /t/ in the axis
        :rtype: int, float
        """
        if self.name == 'time':
            timex = self.get_time_axis(step)
        else:
            timex = self.get_wave(step)
        for i, x in enumerate(timex):
            if x == t:
                return i
            elif x > t:
                # Needs to interpolate the data
                if i == 0:
                    raise IndexError("Time position is lower than t0")
                frac = (t - timex[i - 1]) / (timex[i] - timex[i - 1])
                return i - 1 + frac

    def get_len(self, step: int = 0) -> int:
        """
        Returns the length of the axis.
        
        :param step: Optional parameter the step index.
        :type step: int
        :return: The number of data points
        :rtype: int
        """
        return self.step_offset(step + 1) - self.step_offset(step)

    def __len__(self):
        if self.step_info is None:
            return len(self.data)
        else:
            return self.get_len()

    def __iter__(self):
        assert self.step_info is None, "Iteration can't be used with stepped data. Use get_wave() method."
        return self.data.__iter__()


class TraceRead(DataSet):
    """This class is used to represent a trace. It derives from DataSet and implements the additional methods to
    support STEPed simulations.
    This class is constructed by the get_trace() command.
    Data can be accessed through the [] and len() operators, or by the get_wave() method.
    If numpy is available the get_wave() method will return a numpy array.
    """

    def __init__(self, name, whattype, datalen, axis, numerical_type='real'):
        super().__init__(name, whattype, datalen, numerical_type)
        self.axis = axis

    def get_point(self, n: int, step: int = 0) -> Union[float, complex]:
        """
        Implementation of the [] operator.

        :param n: item in the array
        :type n: int
        :param step: Optional step number
        :type step: int
        :return: float value of the item
        :rtype: float, complex
        """
        if self.axis is None:
            if n != 0:
                return self.data[n]
            else:
                return self.data[step]  # This is for the case of stepped operation point simulation.
        else:
            return self.data[self.axis.step_offset(step) + n]

    def __getitem__(self, item) -> Union[float, complex]:
        """This is only here for compatibility with previous code. """
        assert self.axis is None or self.axis.step_info is None, \
            "Indexing should not be used with stepped data. Use get_point() method"
        return self.data.__getitem__(item)

    def get_wave(self, step: int = 0) -> np.array:
        """
        Returns the data contained in this object. For stepped simulations an argument must be passed specifying the
        step number. If no steps exist, the argument must be left blank.
        To know whether stepped data exist, the user can use the get_raw_property('Flags') method.

        If numpy is available the get_wave() method will return a numpy array.

        :param step: To be used when stepped data exist on the RAW file.
        :type step: int
        :return: a List or numpy array (if installed) containing the data contained in this object.
        :rtype: numpy.array
        """
        if self.axis is None:
            return super().get_wave()
        else:
            if step == 0:
                return self.data[:self.axis.step_offset(1)]
            else:
                return self.data[self.axis.step_offset(step):self.axis.step_offset(step + 1)]

    def get_point_at(self, t, step: int = 0) -> Union[float, complex]:
        """
        Get a point from the trace at the point specified by the /t/ argument.
        If the point doesn't exist on the axis, the data is interpolated using a linear regression between the two
        adjacent points.
        
        :param t: point in the axis where to find the point.
        :type t: float, float32(numpy) or float64(numpy)
        :param step: step index
        :type step: int
        """
        pos = self.axis.get_position(t, step)
        if isinstance(pos, (float, float32, float64)):
            offset = self.axis.step_offset(step)
            i = int(pos)
            last_item = self.get_len(step) - 1
            if i < last_item:
                f = pos - i
                return self.data[offset + i] + f * (self.data[offset + i + 1] - self.data[offset + i])
            elif pos == last_item:  # This covers the case where a float is given containing the last position
                return self.data[offset + i]
            else:
                raise IndexError(f"The highest index is {last_item}. Received {pos}")
        else:
            return self.get_point(pos, step)

    def get_len(self, step: int = 0) -> int:
        """
        Returns the length of the axis.
        
        :param step: Optional parameter the step index.
        :type step: int
        :return: The number of data points
        :rtype: int
        """
        return self.axis.step_offset(step + 1)

    def __len__(self):
        """
        .. deprecated:: 1.0 This is only here for compatibility with previous code.
        """
        assert self.axis is None or self.axis.step_info is None, \
            "len() should not be used with stepped data. Use get_len() method passing the step index"
        return len(self.data)


class DummyTrace(object):
    """Dummy Trace for bypassing traces while reading"""

    def __init__(self, name, whattype, datalen, numerical_type='real'):
        """Base Class for both Axis and Trace Classes.
        Defines the common operations between both."""
        self.name = name
        self.whattype = whattype
        self.datalen = datalen
        self.numerical_type = numerical_type

    def __str__(self):
        return f"name:'{self.name}'\ntype:'{self.whattype}'\nlen:{self.datalen}"


class SpiceReadException(Exception):
    """Custom class for exception handling"""
    ...



# -*- coding: utf-8 -*-

# Convenience direct imports
from .raw.raw_read import RawRead, SpiceReadException
from .raw.raw_write import RawWrite, Trace
from .editor.spice_editor import SpiceEditor, SpiceCircuit
from .editor.asc_editor import AscEditor
from .editor.qsch_editor import QschEditor
from .sim.sim_runner import SimRunner


def all_loggers():
    """
    Returns all the name strings used as logger identifiers.

    :return: A List of strings which contains all the logger's names used in this library.
    :rtype: list[str]
    """
    return [
        "spicelib.AscEditor",
        "spicelib.AsyReader",
        "spicelib.BaseEditor",
        "spicelib.BaseSchematic",
        "spicelib.LTSpiceSimulator",
        "spicelib.LTSteps",
        "spicelib.NGSpiceSimulator",
        "spicelib.qspice_log_reader",
        "spicelib.QSpiceSimulator",
        "spicelib.RawRead",
        "spicelib.RunTask",
        "spicelib.ServerSimRunner",
        "spicelib.SimAnalysis",
        "spicelib.SimClient",
        "spicelib.SimRunner",
        "spicelib.SimServer",
        "spicelib.SimStepper",
        "spicelib.Simulator",
        "spicelib.SpiceEditor",
        "spicelib.Utils",
        "spicelib.XYCESimulator"
    ]


def set_log_level(level):
    """
    Sets the logging level for all loggers used in the library.

    :param level: The logging level to be used, eg. logging.ERROR, logging.DEBUG, etc.
    :type level: int
    """
    import logging
    for logger in all_loggers():
        logging.getLogger(logger).setLevel(level)


def add_log_handler(handler):
    """
    Sets the logging handler for all loggers used in the library.

    :param handler: The logging handler to be used, eg. logging.NullHandler
    :type handler: Handler
    """
    import logging
    for logger in all_loggers():
        logging.getLogger(logger).addHandler(handler)



from spicelib.sim.sim_runner import SimRunner
from spicelib.editor.spice_editor import SpiceEditor
from spicelib.simulators.ngspice_simulator import NGspiceSimulator
from spicelib.utils.sweep_iterators import sweep_log


def processing_data(raw_file, log_file):
    print("Handling the simulation data of %s, log file %s" % (raw_file, log_file))


# select spice model
LTC = SimRunner(output_folder='./temp', simulator=NGspiceSimulator.create_from('C:/Apps/NGSpice64/bin/ngspice.exe'))
netlist = SpiceEditor('./testfiles/testfile_ngspice.net')
# set default arguments
netlist['R1'].value_str = '4k'
netlist['V1'].model = "SINE(0 1 3k 0 0 0)"  # Modifying the behavior of the voltage source
netlist.remove_instruction('.op')
netlist.add_instruction(".tran 1n 3m")
netlist.add_instruction(".plot V(out)")
netlist.add_instruction(".save all")

# .step dec param cap 1p 10u 1
for cap in sweep_log(1e-12, 10e-6, 10):
    netlist['C1'].value = cap
    LTC.run(netlist, callback=processing_data)

LTC.wait_completion()

# Sim Statistics
print('Successful/Total Simulations: ' + str(LTC.okSim) + '/' + str(LTC.runno))



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        sim_runner.py
# Purpose:     Tool used to launch LTSpice simulation in batch mode.
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     23-12-2016
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
"""
Allows launching LTSpice simulations from a Python Script, thus allowing to overcome the 3 dimensions STEP limitation on
LTSpice, update resistor values, or component models.

The code snipped below will simulate a circuit with two different diode models, set the simulation
temperature to 80 degrees, and update the values of R1 and R2 to 3.3k. ::

    from spicelib.sim.sim_runner import SimRunner
    from spicelib.sim.sweep import sweep
    from spicelib.editor.spice_editor import SpiceEditor
    from spicelib.sim.ltspice_simulator import LTspice

    runner = SimRunner(simulator=LTspice, parallel_sims=4)
    editor = SpiceEditor("my_circuit.net")
    editor.set_parameters(temp=80)  # Sets the simulation temperature to be 80 degrees
    editor.set_component_value('R2', '3.3k')  #  Updates the resistor R2 value to be 3.3k
    for dmodel in ("BAT54", "BAT46WJ"):
        editor.set_element_model("D1", model)  # Sets the Diode D1 model
        for res_value in sweep(2.2, 2,4, 0.2):  # Steps from 2.2 to 2.4 with 0.2 increments
            editor.set_component_value('R1', res_value)  #  Updates the resistor R1 value to be 3.3k
            runner.run()

    runner.wait_completion()  # Waits for the LTSpice simulations to complete

    print("Total Simulations: {}".format(runner.runno))
    print("Successful Simulations: {}".format(runner.okSim))
    print("Failed Simulations: {}".format(runner.failSim))

The first line will create a python class instance that represents the LTSpice file or netlist that is to be
simulated. This object implements methods that are used to manipulate the spice netlist. For example, the method
set_parameters() will set or update existing parameters defined in the netlist. The method set_component_value() is
used to update existing component values or models.

---------------
Multiprocessing
---------------

For making better use of today's computer capabilities, the SimRunner spawns several simulation processes
each executing in parallel a simulation.

By default, the number of parallel simulations is 4, however the user can override this in two ways. Either
using the class constructor argument ``parallel_sims`` or by forcing the allocation of more processes in the
run() call by setting ``wait_resource=False``. ::

    `runner.run(wait_resource=False)`

The recommended way is to set the parameter ``parallel_sims`` in the class constructor. ::

    `runner = SimRunner(simulator=LTspice, parallel_sims=8)`

The user then can launch a simulation with the updates done to the netlist by calling the run() method. Since the
processes are not executed right away, but rather just scheduled for simulation, the wait_completion() function is
needed if the user wants to execute code only after the completion of all scheduled simulations.

The usage of wait_completion() is optional. Just note that the script will only end when all the scheduled tasks are
executed.

---------
Callbacks
---------

As seen above, the `wait_completion()` can be used to wait for all the simulations to be finished. However, this is
not efficient from a multiprocessor point of view. Ideally, the post-processing should be also handled while other
simulations are still running. For this purpose, the user can use a function call back.

The callback function is called when the simulation has finished directly by the thread that has handling the
simulation. A function callback receives two arguments.
The RAW file and the LOG file names. Below is an example of a callback function::

    def processing_data(raw_filename, log_filename):
        '''This is a call back function that just prints the filenames'''
        print("Simulation Raw file is %s. The log is %s" % (raw_filename, log_filename)
        # Other code below either using ltsteps.py or raw_read.py
        log_info = LTSpiceLogReader(log_filename)
        log_info.read_measures()
        rise, measures = log_info.dataset["rise_time"]

The callback function is optional. If  no callback function is given, the thread is terminated just after the
simulation is finished.
"""
__author__ = "Nuno Canto Brum <nuno.brum@gmail.com>"
__copyright__ = "Copyright 2020, Fribourg Switzerland"

__all__ = ['SimRunner', 'SimRunnerTimeoutError', 'AnyRunner', 'ProcessCallback', 'RunTask', 'clock_function']

import shutil
import inspect  # Library used to get the arguments of the callback function
from pathlib import Path
from time import sleep, thread_time as clock
from typing import Callable, Union, Type, Protocol, Tuple
import logging

from .process_callback import ProcessCallback
from ..sim.run_task import RunTask, clock_function
from ..sim.simulator import Simulator
from ..editor.base_editor import BaseEditor

_logger = logging.getLogger("spicelib.SimRunner")
END_LINE_TERM = '\n'


class SimRunnerTimeoutError(TimeoutError):
    """Timeout Error class"""
    ...


class AnyRunner(Protocol):
    def run(self, netlist: Union[str, Path, BaseEditor], *,
            wait_resource: bool = True,
            callback: Union[Type[ProcessCallback], Callable] = None,
            callback_args: Union[tuple, dict] = None,
            switches=None,
            timeout: float = None, run_filename: str = None) -> Union[RunTask, None]:
        ...

    def wait_completion(self, timeout=None, abort_all_on_timeout=False) -> bool:
        ...


class SimRunner(object):
    """
    The SimRunner class implements all the methods required for launching batches of Spice simulations.

    :raises FileNotFoundError: When the file is not found  /!\\ This will be changed

    :param parallel_sims: Defines the number of parallel simulations that can be executed at the same time. Ideally this
                          number should be aligned to the number of CPUs (processor cores) available on the machine.
    :type parallel_sims: int, optional
    :param timeout: Timeout parameter as specified on the os subprocess.run() function. Default is 600 seconds, i.e.
        10 minutes. For no timeout, set to None.
    :type timeout: float, optional
    :param verbose: If True, it enables a richer printout of the program execution.
    :type verbose: bool, optional
    :param output_folder: specifying which directory shall be used for simulation files (raw and log files).
    :type output_folder: str, optional
    :param simulator: Forcing a given simulator executable.
    :type simulator: Simulator, optional
    """

    def __init__(self, *, simulator=None, parallel_sims: int = 4, timeout: float = 600.0, verbose=False,
                 output_folder: str = None):
        # The '*' in the parameter list forces the user to use named parameters for the rest of the parameters.
        # This is a good practice to avoid confusion.
        self.verbose = verbose
        self.timeout = timeout
        self.cmdline_switches = []

        if output_folder:
            self.output_folder = Path(output_folder)  # If not None converts to Path() object
            if not self.output_folder.exists():
                self.output_folder.mkdir()
        else:
            self.output_folder = None

        self.parallel_sims = parallel_sims
        self.active_tasks = []
        self.completed_tasks = []
        self._iterator_counter = 0  # Note: Nested iterators are not supported

        self.runno = 0  # number of total runs
        self.failSim = 0  # number of failed simulations
        self.okSim = 0  # number of successful completed simulations
        # self.failParam = []  # collects for later user investigation of failed parameter sets

        # Gets a simulator.
        if simulator is None:
            raise ValueError("No default simulator defined, please specify a simulator")
        elif issubclass(simulator, Simulator):
            self.simulator = simulator
        else:
            raise TypeError("Invalid simulator type.")
        _logger.info("SimRunner initialized")

    def __del__(self):
        """Class Destructor : Closes Everything"""
        # _logger.debug("Waiting for all spawned sim_tasks to finish.")
        self.wait_completion(abort_all_on_timeout=True)  # Kill all pending simulations
        # _logger.debug("Exiting SimRunner")

    def set_simulator(self, spice_tool: Type[Simulator]) -> None:
        """
        Manually overriding the simulator to be used.

        :param spice_tool: String containing the path to the spice tool to be used, or alternatively the Simulator
            object.
        :type spice_tool: Simulator type
        :return: Nothing
        """
        if issubclass(spice_tool, Simulator):
            self.simulator = spice_tool
        else:
            raise TypeError("Expecting str or Simulator objects")

    def clear_command_line_switches(self):
        """Clear all the command line switches added previously"""
        self.cmdline_switches.clear()

    def add_command_line_switch(self, switch, path=''):
        """
        Used to add an extra command line argument such as -I<path> to add symbol search path or -FastAccess
        to convert the raw file into Fast Access.
        The argument is a string as is defined in the LTSpice command line documentation.

        :param switch: switch to be added.
        :type switch: str:  A command line switch such as "-ascii" for generating a raw file in text format or
            "-alt" for setting the solver to alternate. See Command Line Switches information on LTSpice help file.
        :param path: path to the file related to the switch being given.
        :type path: str, optional
        :returns: Nothing
        """
        self.cmdline_switches.append(switch)
        if path is not None:
            self.cmdline_switches.append(path)

    def _on_output_folder(self, afile):
        if self.output_folder:
            return self.output_folder / Path(afile).name
        else:
            return Path(afile)

    def _to_output_folder(self, afile: Path, *, copy: bool, new_name: str = ''):
        if self.output_folder:
            if new_name:
                ddst = self.output_folder / new_name
            else:
                ddst = self.output_folder

            if copy:
                dest = shutil.copy(afile, ddst)
            else:
                dest = shutil.move(afile, ddst)
            return Path(dest)
        else:
            if new_name:
                dest = shutil.copy(afile, afile.parent / new_name)
                return Path(dest)
            else:
                return afile

    def _run_file_name(self, netlist):
        if not isinstance(netlist, Path):
            netlist = Path(netlist)
        if netlist.suffix == '.qsch':
            # The Qsch files can't be simulated, so, they have to be converted to netlist first.
            netlist = netlist.with_suffix('.net')
        return "%s_%i%s" % (netlist.stem, self.runno, netlist.suffix)

    def _prepare_sim(self, netlist: Union[str, Path, BaseEditor], run_filename: str):
        """Internal function"""
        # update number of simulation
        self.runno += 1  # Incrementing internal simulation number
        # Harmonize the netlist into a Path object pointing to a netlist file on the right output folder
        if isinstance(netlist, BaseEditor):
            if run_filename is None:
                run_filename = self._run_file_name(netlist.circuit_file)

            # Calculates the path where to store the new netlist.
            run_netlist_file = self._on_output_folder(run_filename)
            netlist.save_netlist(run_netlist_file)

        elif isinstance(netlist, (Path, str)):
            if run_filename is None:
                run_filename = self._run_file_name(netlist)
            if isinstance(netlist, str):
                netlist = Path(netlist)
            run_netlist_file = self._to_output_folder(netlist, copy=True, new_name=run_filename)
        else:
            raise TypeError("'netlist' parameter shall be a SpiceEditor, pathlib.Path or a plain str")

        return run_netlist_file

    @staticmethod
    def validate_callback_args(callback: Callable, callback_args: Union[tuple, dict]) -> Union[dict, None]:
        """
        It validates that the callback_args are matching the callback function.
        Note that the first two parameters of the callback functions need to be the raw and log files.

        """
        if callback is None:
            return None  # No callback function, hence callback_args have no effect
        if inspect.isclass(callback) and issubclass(callback, ProcessCallback):
            args = inspect.signature(callback.callback).parameters
        else:
            args = inspect.signature(callback).parameters
        if len(args) < 2:
            raise ValueError("Callback function must have at least two arguments")
        if len(args) > 2:
            if callback_args is None:
                raise ValueError("Callback function has more than two arguments, but no callback_args are given")
            if isinstance(callback_args, dict):
                for pos, param in enumerate(args):
                    if pos > 1:
                        if param not in callback_args:
                            raise ValueError("Callback argument '%s' not found in callback_args" % param)

            if len(args) - 2 != len(callback_args):
                raise ValueError("Callback function has %d arguments, but %d callback_args are given" %
                                 (len(args), len(callback_args))
                                 )
            if isinstance(callback_args, tuple):
                # Convert into a dictionary
                return {param: callback_args[pos - 2] for pos, param in enumerate(args) if pos > 1}
            else:
                return callback_args

    def run(self, netlist: Union[str, Path, BaseEditor], *, wait_resource: bool = True,
            callback: Union[Type[ProcessCallback], Callable] = None,
            callback_args: Union[tuple, dict] = None,
            switches=None,
            timeout: float = None, run_filename: str = None) -> Union[RunTask, None]:
        """
        Executes a simulation run with the conditions set by the user.
        Conditions are set by the set_parameter, set_component_value or add_instruction functions.

        :param netlist:
            The name of the netlist can be optionally overridden if the user wants to have a better control of how the
            simulations files are generated.
        :type netlist: SpiceEditor or a path to the file
        :param wait_resource:
            Setting this parameter to False will force the simulation to start immediately, irrespective of the number
            of simulations already active.
            By default, the SimRunner class uses only four processors. This number can be overridden by setting
            the parameter ´parallel_sims´ to a different number.
            If there are more than ´parallel_sims´ simulations being done, the new one will be placed on hold till one
            of the other simulations are finished.
        :type wait_resource: bool, optional
        :param callback:
            The user can optionally give a callback function for when the simulation finishes so that processing can
            be done immediately. The callback can either be a function or a class derived from ProcessCallback.
            A callback function must receive two at least input parameters that correspond the
            raw and log files created by the simulation. These need to be the first two parameters of the callback
            function. The other parameters are passed as a dictionary or a tuple in the callback_args parameter.
            If the callback is a class derived from ProcessCallback, then the callback is executed in a separate
            process. The callback function must be defined in the callback() method of the class. As for the callback
            function, the first two parameters are the raw and log files. The other parameters are passed as dictionary
            in the callback_args parameter.

        :type: callback: function(raw_file: Path, log_file: Path, ...), optional
        :param callback_args:
            The callback function arguments. This parameter is passed as keyword arguments to the callback function.
        :type callback_args: dict or tuple, optional
        :param switches: Command line switches override
        :type switches: list
        :param timeout:
            Timeout to be used in waiting for resources. Default time is value defined in this class constructor.
        :type timeout: float, optional
        :param run_filename: Name to be used for the log and raw file.
        :type run_filename: str or Path
        :returns: The task object of type RunTask
        """
        callback_kwargs = self.validate_callback_args(callback, callback_args)
        if switches is None:
            switches = []
        run_netlist_file = self._prepare_sim(netlist, run_filename)

        if timeout is None:
            timeout = self.timeout

        t0 = clock()  # Store the time for timeout calculation
        while clock() - t0 < timeout + 1:  # Give one second slack in relation to the task timeout
            cmdline_switches = switches or self.cmdline_switches  # If switches are passed, they override the ones
            # inside the class.

            if (wait_resource is False) or (self.active_threads() < self.parallel_sims):
                t = RunTask(self.simulator, self.runno, run_netlist_file, callback, callback_kwargs,
                            cmdline_switches, timeout=timeout, verbose=self.verbose)
                self.active_tasks.append(t)
                t.start()
                sleep(0.01)  # Give slack for the thread to start
                return t  # Returns the task object
            sleep(0.1)  # Give Time for other simulations to end
        else:
            _logger.error("Timeout waiting for resources for simulation %d" % self.runno)
            if self.verbose:
                _logger.warning("Timeout on launching simulation %d." % self.runno)
            return None

    def run_now(self, netlist: Union[str, Path, BaseEditor], *, switches=None, run_filename: str = None,
                timeout: float = None) -> Tuple[str, str]:
        """
        Executes a simulation run with the conditions set by the user.
        Conditions are set by the set_parameter, set_component_value or add_instruction functions.

        :param netlist:
            The name of the netlist can be optionally overridden if the user wants to have a better control of how the
            simulations files are generated.
        :type netlist: SpiceEditor or a path to the file
        :param switches: Command line switches override
        :type switches: list
        :param run_filename: Name to be used for the log and raw file.
        :type run_filename: str or Path
        :param timeout: Timeout to be used in waiting for resources. Default time is value defined in this class
            constructor.
        :type timeout: float, optional
        :returns: the raw and log filenames
        """
        if switches is None:
            switches = []
        run_netlist_file = self._prepare_sim(netlist, run_filename)

        cmdline_switches = switches or self.cmdline_switches  # If switches are passed, they override the ones inside
        # the class.

        if timeout is None:
            timeout = self.timeout

        def dummy_callback(raw, log):
            """Dummy call back that does nothing"""
            return None

        t = RunTask(
            simulator=self.simulator, runno=self.runno, netlist_file=run_netlist_file,
            callback=dummy_callback, callback_args=None,
            switches=cmdline_switches, timeout=timeout, verbose=self.verbose
        )
        t.start()
        sleep(0.01)  # Give slack for the thread to start
        t.join(timeout + 1)  # Give one second slack in relation to the task timeout
        self.completed_tasks.append(t)
        if t.retcode == 0:
            self.okSim += 1
        else:
            # simulation failed
            self.failSim += 1
        return t.raw_file, t.log_file  # Returns the raw and log file

    def active_threads(self):
        """Returns the number of active sim_tasks"""
        self.update_completed()
        return len(self.active_tasks)

    def update_completed(self):
        """
        This function updates the active_tasks and completed_tasks lists. It moves the finished task from the
        active_tasks list to the completed_tasks list.
        It should be called periodically to update the status of the simulations.

        :returns: Nothing
        """
        i = 0
        while i < len(self.active_tasks):
            if self.active_tasks[i].is_alive():
                i += 1
            else:
                if self.active_tasks[i].retcode == 0:
                    self.okSim += 1
                else:
                    # simulation failed
                    self.failSim += 1
                task = self.active_tasks.pop(i)
                self.completed_tasks.append(task)

    def kill_all_ltspice(self):
        """
        .. deprecated:: 1.0 Use `kill_all_spice()` instead.
        
        This is only here for compatibility with previous code.
        
        Function to terminate LTSpice"""
        self.kill_all_spice()
        
    def kill_all_spice(self):
        """Function to terminate xxSpice processes"""
        simulator = Simulator
        process_name = simulator.process_name
        import psutil
        for proc in psutil.process_iter():
            # check whether the process name matches

            if proc.name() == process_name:
                _logger.info("killing Spice", proc.pid)
                proc.kill()

    def _maximum_stop_time(self):
        """
        This function will return the maximum timeout time of all active tasks.
        :return: Maximum timeout time or None, if there is no timeout defined.
        :rtype: float or None
        """
        alarm = None
        for task in self.active_tasks:
            tout = task.timeout if task.timeout is not None else self.timeout
            if tout is not None:
                stop = task.start_time + tout
                if alarm is None:
                    alarm = stop
                elif stop > alarm:
                    alarm = stop
        return alarm

    def wait_completion(self, timeout=None, abort_all_on_timeout=False) -> bool:
        """
        This function will wait for the execution of all scheduled simulations to complete.

        :param timeout: Cancels the wait after the number of seconds specified by the timeout.
            This timeout is reset everytime that a simulation is completed. The difference between this timeout and the
            one defined in the SimRunner instance, is that the latter is implemented by the subprocess class, and
            this one just cancels the wait.
        :type timeout: int
        :param abort_all_on_timeout: attempts to stop all LTSpice processes if timeout is expired.
        :type abort_all_on_timeout: bool
        :returns: True if all simulations were executed successfully
        :rtype: bool
        """
        self.update_completed()
        if timeout is not None:
            stop_time = clock_function() + timeout
        while len(self.active_tasks) > 0:
            sleep(1)
            self.update_completed()
            if timeout is None:
                stop_time = self._maximum_stop_time()
            if stop_time is not None:  # This can happen if timeout was set as none everywhere
                if clock_function() > stop_time:
                    if abort_all_on_timeout:
                        self.kill_all_spice()
                    return False

        return self.failSim == 0

    @staticmethod
    def _del_file_if_exists(workfile: Path):
        """
        Deletes a file if it exists.
        :param workfile: File to be deleted
        :type workfile: Path
        :return: Nothing
        """
        if workfile is not None and workfile.exists():
            _logger.info("Deleting..." + workfile.name)
            workfile.unlink()

    @staticmethod
    def _del_file_ext_if_exists(workfile: Path, ext: str):
        """
        Deletes a file extension if it exists.
        :param workfile: File to be deleted
        :type workfile: Path
        :param ext: Extension to be deleted
        :type ext: str
        :return: Nothing
        """
        sim_file = workfile.with_suffix(ext)
        SimRunner._del_file_if_exists(sim_file)

    def cleanup_files(self):
        """
        Will delete all log and raw files that were created by the script. This should only be executed at the end
        of data processing.
        """
        self.update_completed()  # Updates the active_tasks and completed_tasks lists

        for task in self.completed_tasks:
            netlistfile = task.netlist_file
            self._del_file_if_exists(netlistfile)  # Delete the netlist file if still exists
            self._del_file_if_exists(task.log_file)  # Delete the log file if was created
            self._del_file_if_exists(task.raw_file)  # Delete the raw file if was created

            if netlistfile.suffix == '.net' or netlistfile.suffix == '.asc':
                # Delete the files that have been potentially created by LTSpice
                for ext in ('.log.raw', '.op.raw'):
                    self._del_file_ext_if_exists(netlistfile, ext)

                if netlistfile.suffix == '.asc':  # If simulated from an asc file, delete the .net file
                    # Then needs to delete the .net as well
                    self._del_file_ext_if_exists(netlistfile, '.net')

    def file_cleanup(self):
        """
        .. deprecated:: 1.0 Use `cleanup_files()` instead.
        """
        self.cleanup_files()  # Alias for backward compatibility, this will be deleted in the future

    def __iter__(self):
        self._iterator_counter = 0  # Reset the iterator counter. Note: nested iterators are not supported
        return self

    def __next__(self):
        while True:
            self.update_completed()  # Updates the active_tasks and completed_tasks lists
            # First go through the completed tasks
            if self._iterator_counter < len(self.completed_tasks):
                ret = self.completed_tasks[self._iterator_counter]
                self._iterator_counter += 1
                return ret.get_results()

            # Then check if there are any active tasks
            if len(self.active_tasks) == 0:
                raise StopIteration

            # Then go through the active tasks to get the maximum timeout
            stop_time = self._maximum_stop_time()

            if stop_time is not None and clock_function() > stop_time:  # All tasks are on timeout condition
                raise SimRunnerTimeoutError(f"Exceeded {self.timeout} seconds waiting for tasks to finish")

            # Wait for the active tasks to finish with a timeout
            sleep(0.2)  # Go asleep for a while



# coding=utf-8
# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        base_editor.py
# Purpose:     Abstract class that defines the protocol for the editors
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------

__author__ = "Nuno Canto Brum <nuno.brum@gmail.com>"
__version__ = "0.1.0"
__copyright__ = "Copyright 2021, Fribourg Switzerland"

from abc import ABC, abstractmethod
from collections import OrderedDict
from math import floor, log
from pathlib import Path
from typing import Union, List
import logging
from ..sim.simulator import Simulator


_logger = logging.getLogger("spicelib.BaseEditor")

SUBCKT_DIVIDER = ':'  #: This controls the sub-circuit divider when setting component values inside sub-circuits.
# Ex: Editor.set_component_value('XU1:R1', '1k')

UNIQUE_SIMULATION_DOT_INSTRUCTIONS = ('.AC', '.DC', '.TRAN', '.NOISE', '.DC', '.TF')
SPICE_DOT_INSTRUCTIONS = (
    '.BACKANNO',
    '.END',
    '.ENDS',
    '.FERRET',  # Downloads a File from a given URL
    '.FOUR',  # Compute a Fourier Component after a .TRAN Analysis
    '.FUNC', '.FUNCTION',
    '.GLOBAL',
    '.IC',
    '.INC', '.INCLUDE',  # Include another file
    '.LIB',  # Include a Library
    '.LOADBIAS',  # Load a Previously Solved DC Solution
    # These Commands are part of the contraption Programming Language of the Arbitrary State Machine
    '.MACHINE', '.STATE', '.RULE', '.OUTPUT', '.ENDMACHINE',
    '.MEAS', '.MEASURE',
    '.MODEL',
    '.NET',  # Compute Network Parameters in a .AC Analysis
    '.NODESET',  # Hints for Initial DC Solution
    '.OP',
    '.OPTIONS',
    '.PARAM', '.PARAMS',
    '.SAVE', '.SAV',
    '.SAVEBIAS',
    '.STEP',
    '.SUBCKT',
    '.TEXT',
    '.WAVE',  # Write Selected Nodes to a .Wav File

)


def PARAM_REGEX(pname):
    return r"(?P<name>" + pname + r")\s*[= ]\s*(?P<value>[\w\*\/\.\+\-\/\*\{\}\(\)%]*)"


def format_eng(value) -> str:
    """
    Helper function for formatting value with the SI qualifiers.  That is, it will use

        * p for pico (10E-12)
        * n for nano (10E-9)
        * u for micro (10E-6)
        * m for mili (10E-3)
        * k for kilo (10E+3)
        * Meg for Mega (10E+6)
        * g for giga (10E+9)
        * t for tera (10E+12)


    :param value: float value to format
    :type value: float
    :return: String with the formatted value
    :rtype: str
    """
    if value == 0.0:
        return "{:g}".format(value)  # This avoids a problematic log(0), and the int and float conversions
    e = floor(log(abs(value), 1000))
    if -5 <= e < 0:
        suffix = "fpnum"[e]
    elif e == 0:
        return "{:g}".format(value)
    elif e == 1:
        suffix = "k"
    elif e == 2:
        suffix = 'Meg'
    elif e == 3:
        suffix = 'g'
    elif e == 4:
        suffix = 't'
    else:
        return '{:E}'.format(value)
    return '{:g}{:}'.format(value * 1000 ** -e, suffix)


def scan_eng(value: str) -> float:
    """
    Converts a string to a float, considering SI multipliers

        * f for femto (10E-15)
        * p for pico (10E-12)
        * n for nano (10E-9)
        * u or µ for micro (10E-6)
        * m for mili (10E-3)
        * k for kilo (10E+3)
        * meg for Mega (10E+6)
        * g for giga (10E+9)
        * t for tera (10E+12)
        
    The extra unit qualifiers such as V for volts or F for Farads are ignored.


    :param value: string to be converted to float
    :type value: str
    :return:
    :rtype: float
    :raises: ValueError when the value cannot be converted.
    """
    # Search for the last digit on the string. Assuming that all after the last number are SI qualifiers and units.
    value = value.strip()
    x = len(value)
    while x > 0:
        if value[x - 1] in "0123456789":
            break
        x -= 1
    suffix = value[x:]  # this is the non-numeric part at the end
    f = float(value[:x])  # this is the numeric part. Can raise ValueError.
    if suffix:
        suffix = suffix.lower()
        # By industry convention, SPICE is not case sensitive
        if suffix.startswith("meg"):
            return f * 1E+6
        elif suffix[0] in "fpnuµmkgt":
            return f * {
                'f': 1.0e-15,
                'p': 1.0e-12,
                'n': 1.0e-09,
                'u': 1.0e-06,
                'µ': 1.0e-06,
                'm': 1.0e-03,
                'k': 1.0e+03,
                'g': 1.0e+09,
                't': 1.0e+12,
            }[suffix[0]]
    return f


def to_float(value, accept_invalid: bool = True) -> (float, str):
    _MULT = {
        'f': 1E-15,
        'p': 1E-12,
        'n': 1E-9,
        'u': 1E-6,
        'U': 1E-6,
        'm': 1E-3,
        'M': 1E-3,
        'k': 1E+3,
        'K': 1E+3,  # For much of the world, K is the same as k. That is a sad fact of life. K is Kelvin in SI
        'Meg': 1E+6,
        'g': 1E+9,
        't': 1E+12,
        # These units can be used as decimal points in the number definition. Ex: 10R5 is 10.5 Ohms. In LTSpice
        # the units can be used in any number definition. For example 10H5 is 10.5 Henrys but also can be used in
        # resistors value definition. LTSpice doesn't care about the unit in the component value definition.
        'Ω': 1,  # This is the Ohm symbol. It is supported by LTspice
        'R': 1,  # This also represents the Ohm symbol. Can be used a decimal point. Ex: 10R2 is 10.2 Ohms
        'V': 1,  # Volts
        'A': 1,  # Amperes (Current)
        'F': 1,  # Farads (Capacitance)
        'H': 1,  # Henry (Inductance)
        '%': 0.01,  # Percent. 10% is 0.1. 1%6 is 0.016
    }

    value = value.strip()  # Removing trailing and leading spaces
    length = len(value)

    multiplier = 1.0

    i = 0
    while i < length and (value[i] in "0123456789.+-"):  # Includes spaces
        i += 1
    if i == 0:
        if accept_invalid:
            return value
        else:
            raise ValueError("Doesn't start with a number")

    if 0 < i < length and (value[i] == 'E' or value[i] == 'e'):
        # if it is a number in scientific format, it doesn't have 1000x qualifiers (Ex: p, u, k, etc...)
        i += 1
        while i < length and (value[i] in "0123456789+-"):  # Includes spaces
            i += 1
        j = k = i
    else:
        # this first part should be able to be converted into float
        k = i  # Stores the position of the end of the number
        # Consume any spaces that may exist between the number and the unit
        while i < length and (value[i] in " \t"):
            i += 1

        if i < length:  # Still has characters to consume
            if value[i] in _MULT:
                if value[i:].upper().startswith('MEG'):  # to 1E+06 qualifier 'Meg'
                    i += 3
                    multiplier = _MULT['Meg']
                else:
                    multiplier = _MULT[value[i]]
                    i += 1

            # This part is done to support numbers with the format 1k7 or 1R8
            j = i
            while i < length and (value[i] in "0123456789"):
                i += 1
        else:
            j = i

    try:
        if j < i:  # There is a suffix number
            value = float(value[:k] + "." + value[j:i]) * multiplier
        else:
            value = float(value[:k]) * multiplier
    except ValueError as err:
        if not accept_invalid:
            raise err
    return value


class ComponentNotFoundError(Exception):
    """Component Not Found Error"""


class ParameterNotFoundError(Exception):
    """ParameterNotFound Error"""

    def __init__(self, parameter):
        super().__init__(f'Parameter "{parameter}" not found')


class Primitive(object):
    """Holds the information of a primitive element in the netlist. This is a base class for the Component and is
    used to hold the information of the netlist primitives, such as .PARAM, .OPTIONS, .IC, .NODESET, .GLOBAL, etc.
    """

    def __init__(self, line: str):
        self.line = line

    def append(self, line):
        self.line += line

    def __str__(self):
        return self.line


class Component(Primitive):
    """Holds component information"""

    def __init__(self, parent, line: str):
        super().__init__(line)
        self.reference = ""
        self.attributes = OrderedDict()
        self.ports = []
        self.parent = parent

    @property
    def value_str(self):
        return self.parent.get_component_value(self.reference)

    @value_str.setter
    def value_str(self, value):
        self.parent.set_component_value(self.reference, value)

    @property
    def params(self) -> OrderedDict:
        return self.parent.get_component_parameters(self.reference)

    def set_params(self, **param_dict):
        self.parent.set_component_parameters(self.reference, **param_dict)

    @property
    def value(self):
        return to_float(self.value_str, accept_invalid=True)

    @property
    def model(self):
        return self.parent.get_element_value(self.reference)

    @model.setter
    def model(self, model):
        self.parent.set_element_model(self.reference, model)

    @value.setter
    def value(self, value):
        if isinstance(value, (int, float)):
            self.value_str = format_eng(value)
        else:
            self.value_str = value

    def __str__(self):
        return f"{self.reference} = {self.value}"

    def __getitem__(self, item):
        return self.attributes[item]

    def __setitem__(self, key, value):
        self.attributes[key] = value


class BaseEditor(ABC):
    """
    This defines the primitives (protocol) to be used for both SpiceEditor and AscEditor
    classes.
    """
    custom_lib_paths: List[str] = []  # This is a class variable, so it will be shared between all instances.
    """:meta hide-value:"""    
    simulator_lib_paths: List[str] = []  # This is a class variable, so it will be shared between all instances.
    """:meta hide-value:"""

    @property
    @abstractmethod
    def circuit_file(self) -> Path:
        """Returns the path of the circuit file."""
        ...

    @abstractmethod
    def reset_netlist(self, create_blank: bool = False) -> None:
        """
        Reverts all changes done to the netlist. If create_blank is set to True, then the netlist is blanked.

        :param create_blank: If True, the netlist will be reset to a new empty netlist. If False, the netlist will be
                             reset to the original state.
        """
        ...

    @abstractmethod
    def save_netlist(self, run_netlist_file: Union[str, Path]) -> None:
        """
        Saves the current state of the netlist to a file.

        :param run_netlist_file: File name of the netlist file.
        :type run_netlist_file: Path or str
        :returns: Nothing
        """
        ...

    def write_netlist(self, run_netlist_file: Union[str, Path]) -> None:
        """
        .. deprecated:: 1.x Use `save_netlist()` instead.

        Writes the netlist to a file. This is an alias to save_netlist."""
        self.save_netlist(run_netlist_file)

    @abstractmethod
    def get_component(self, reference: str) -> Component:
        """Returns the Component object representing the given reference in the netlist."""
        ...

    @abstractmethod
    def get_subcircuit(self, reference: str) -> 'BaseEditor':
        """Returns a hierarchical subdesign"""
        ...

    def __getitem__(self, item) -> Component:
        """
        This method allows the user to get the value of a component using the syntax:
        component = circuit['R1']
        """
        return self.get_component(item)

    def __setitem__(self, key, value):
        self.set_component_value(key, value)

    def get_component_attribute(self, reference: str, attribute: str) -> str:
        """Returns the value of the attribute of the component. Attributes are the values that are not related with
        SPICE parameters. For example, component manufacturer, footprint, schematic appearance, etc.
        User can define whatever attributes they want. The only restriction is that the attribute name must be a string.
        :param reference: Reference of the component
        :type reference: str
        :param attribute: Name of the attribute to be retrieved
        :type attribute: str
        :return: Value of the attribute being sought
        :rtype: str
        :raises: ComponentNotFoundError - In case the component is not found
                 KeyError - In case the attribute is not found
        """
        return self.get_component(reference).attributes[attribute]

    def get_component_nodes(self, reference: str) -> list:
        """Returns the value of the port of the component.
        :param reference: Reference of the component
        :type reference: str
        :return: List with the ports of the component
        :rtype: str
        :raises: ComponentNotFoundError - In case the component is not found
                 KeyError - In case the port is not found
        """
        return self.get_component(reference).ports

    @abstractmethod
    def get_parameter(self, param: str) -> str:
        """
        Retrieves a Parameter from the Netlist

        :param param: Name of the parameter to be retrieved
        :type param: str
        :return: Value of the parameter being sought
        :rtype: str
        :raises: ParameterNotFoundError - In case the component is not found
        """
        ...

    def set_parameter(self, param: str, value: Union[str, int, float]) -> None:
        """Adds a parameter to the SPICE netlist.

        Usage: ::

         editor.set_parameter("TEMP", 80)

        This adds onto the netlist the following line: ::

         .PARAM TEMP=80

        This is an alternative to the set_parameters which is more pythonic in it's usage,
        and allows setting more than one parameter at once.

        :param param: Spice Parameter name to be added or updated.
        :type param: str

        :param value: Parameter Value to be set.
        :type value: str, int or float

        :return: Nothing
        """
        ...

    def set_parameters(self, **kwargs):
        """Adds one or more parameters to the netlist.
        Usage: ::

            for temp in (-40, 25, 125):
                for freq in sweep_log(1, 100E3,):
                    editor.set_parameters(TEMP=80, freq=freq)

        :key param_name:
            Key is the parameter to be set. values the ther corresponding values. Values can either be a str; an int or
            a float.

        :returns: Nothing
        """
        for param in kwargs:
            self.set_parameter(param, kwargs[param])

    @abstractmethod
    def set_component_value(self, device: str, value: Union[str, int, float]) -> None:
        """Changes the value of a component, such as a Resistor, Capacitor or Inductor. For components inside
        sub-circuits, use the sub-circuit designator prefix with ':' as separator (Example X1:R1)
        Usage: ::

            editor.set_component_value('R1', '3.3k')
            editor.set_component_value('X1:C1', '10u')

        :param device: Reference of the circuit element to be updated.
        :type device: str
        :param value:
            value to be set on the given circuit element. Float and integer values will automatically
            formated as per the engineering notations 'k' for kilo, 'm', for mili and so on.
        :type value: str, int or float
        :raises:
            ComponentNotFoundError - In case the component is not found

            ValueError - In case the value doesn't correspond to the expected format

            NotImplementedError - In case the circuit element is defined in a format which is not supported by this
            version.

            If this is the case, use GitHub to start a ticket.  https://github.com/nunobrum/spicelib
        """
        ...

    @abstractmethod
    def set_element_model(self, element: str, model: str) -> None:
        """Changes the value of a circuit element, such as a diode model or a voltage supply.
        Usage: ::

            editor.set_element_model('D1', '1N4148')
            editor.set_element_model('V1' "SINE(0 1 3k 0 0 0)")

        :param element: Reference of the circuit element to be updated.
        :type element: str
        :param model: model name of the device to be updated
        :type model: str

        :raises:
            ComponentNotFoundError - In case the component is not found

            ValueError - In case the model format contains irregular characters

            NotImplementedError - In case the circuit element is defined in a format which is not supported by this version.

            If this is the case, use GitHub to start a ticket.  https://github.com/nunobrum/spicelib
        """
        ...

    @abstractmethod
    def set_component_parameters(self, element: str, **kwargs) -> None:
        """
        Adds one or more parameters to the component on the netlist. The argument is in the form of a key-value pair
        where each parameter is the key and the value is value to be set in the netlist.

        Usage 1: ::

         editor.set_component_parameters(R1, value=330, temp=25)

        Usage 2: ::

         value_settings = {'value': 330, 'temp': 25}
         editor.set_component_parameters(R1, **value_settings)

        :param element: Reference of the circuit element.
        :type element: str

        :key <param_name>:
            The key is the parameter name and the value is the value to be set. Values can either be
            strings; integers or floats. When None is given, the parameter will be removed, if possible.

        :return: Nothing
        :raises: ComponentNotFoundError - In case one of the component is not found.
        """
        ...

    def set_component_attribute(self, reference: str, attribute: str, value: str) -> None:
        """Sets the value of the attribute of the component. Attributes are the values that are not related with
        SPICE parameters. For example, component manufacturer, footprint, schematic appearance, etc.
        User can define whatever attributes they want. The only restriction is that the attribute name must be a string.
        :param reference: Reference of the component
        :type reference: str
        :param attribute: Name of the attribute to be set
        :type attribute: str
        :param value: Value of the attribute to be set
        :type value: str
        :return: Nothing
        :raises: ComponentNotFoundError - In case the component is not found
        """
        self.get_component(reference).attributes[attribute] = value

    @abstractmethod
    def get_component_value(self, element: str) -> str:
        """
        Returns the value of a component retrieved from the netlist.

        :param element: Reference of the circuit element to get the value.
        :type element: str

        :return: value of the circuit element .
        :rtype: str

        :raises: ComponentNotFoundError - In case the component is not found

                 NotImplementedError - for not supported operations
        """
        ...

    @abstractmethod
    def get_component_parameters(self, element: str) -> dict:
        """
        Returns the parameters of a component retrieved from the netlist.

        :param element: Reference of the circuit element to get the parameters.
        :type element: str

        :return: parameters of the circuit element in dictionary format.
        :rtype: dict

        :raises: ComponentNotFoundError - In case the component is not found

                 NotImplementedError - for not supported operations
        """
        ...

    def get_component_floatvalue(self, element: str) -> float:
        """
        Returns the value of a component retrieved from the netlist.

        :param element: Reference of the circuit element to get the value in float format.
        :type element: str

        :return: value of the circuit element in float type
        :rtype: float

        :raises: ComponentNotFoundError - In case the component is not found

                 NotImplementedError - for not supported operations
        """
        return scan_eng(self.get_component_value(element))

    def set_component_values(self, **kwargs):
        """
        Adds one or more components on the netlist. The argument is in the form of a key-value pair where each
        component designator is the key and the value is value to be set in the netlist.

        Usage 1: ::

         editor.set_component_values(R1=330, R2="3.3k", R3="1Meg", V1="PWL(0 1 30m 1 30.001m 0 60m 0 60.001m 1)")

        Usage 2: ::

         value_settings = {'R1': 330, 'R2': '3.3k', 'R3': "1Meg", 'V1': 'PWL(0 1 30m 1 30.001m 0 60m 0 60.001m 1)'}
         editor.set_component_values(**value_settings)

        :key <comp_ref>:
            The key is the component designator (Ex: V1) and the value is the value to be set. Values can either be
            strings; integers or floats

        :return: Nothing
        :raises: ComponentNotFoundError - In case one of the component is not found.
        """
        for value in kwargs:
            self.set_component_value(value, kwargs[value])

    @abstractmethod
    def get_components(self, prefixes='*') -> list:
        """
        Returns a list of components that match the list of prefixes indicated on the parameter prefixes.
        In case prefixes is left empty, it returns all the ones that are defined by the REPLACE_REGEXES.
        The list will contain the designators of all components found.

        :param prefixes:
            Type of prefixes to search for. Examples: 'C' for capacitors; 'R' for Resistors; etc... See prefixes
            in SPICE documentation for more details.
            The default prefix is '*' which is a special case that returns all components.
        :type prefixes: str

        :return:
            A list of components matching the prefixes demanded.
        """
        ...

    @abstractmethod
    def add_component(self, component: Component, **kwargs) -> None:
        """
        Adds a component to the design. If the component already exists, it will be replaced by the new one.
        kwargs are implementation specific and can be used to pass additional information to the implementation.

        :param component: Component to be added to the design.
        :type component: Component

        :return: Nothing
        """
        ...

    @abstractmethod
    def remove_component(self, designator: str) -> None:
        """
        Removes a component from  the design.
        Note: Current implementation only allows removal of a component from the main netlist, not from a sub-circuit.

        :param designator: Component reference in the design. Ex: V1, C1, R1, etc...
        :type designator: str

        :return: Nothing
        :raises: ComponentNotFoundError - When the component doesn't exist on the netlist.
        """
        ...

    @abstractmethod
    def add_instruction(self, instruction: str) -> None:
        """
        Adds a SPICE instruction to the netlist.

        For example:

            .. code-block:: text

                .tran 10m ; makes a transient simulation
                .meas TRAN Icurr AVG I(Rs1) TRIG time=1.5ms TARG time=2.5ms" ; Establishes a measuring
                .step run 1 100, 1 ; makes the simulation run 100 times


        :param instruction:
            Spice instruction to add to the netlist. This instruction will be added at the end of the netlist,
            typically just before the .BACKANNO statement
        :type instruction: str
        :return: Nothing
        """
        ...

    @abstractmethod
    def remove_instruction(self, instruction) -> None:
        """
        Removes a SPICE instruction from the netlist.

        Example:

        .. code-block:: python

            editor.remove_instruction(".STEP run -1 1023 1")

        This only works if the instruction exactly matches the line on the netlist. This means that space characters,
        and upper case and lower case differences will not match the line.

        :param instruction: The list of instructions to remove. Each instruction is of the type 'str'
        :type instruction: str
        :returns: Nothing
        """
        ...

    @abstractmethod
    def remove_Xinstruction(self, search_pattern: str) -> None:
        """
        Removes a SPICE instruction from the netlist based on a search pattern. This is a more flexible way to remove
        instructions from the netlist. The search pattern is a regular expression that will be used to match the
        instructions to be removed. The search pattern will be applied to each line of the netlist and if the pattern
        matches, the line will be removed.

        Example: The code below will remove all AC analysis instructions from the netlist.

        .. code-block:: python

            editor.remove_Xinstruction("\\.AC.*")

        :param search_pattern: The list of instructions to remove. Each instruction is of the type 'str'
        :type search_pattern: str
        :returns: Nothing
        """
        ...

    def add_instructions(self, *instructions) -> None:
        """
        Adds a list of instructions to the SPICE NETLIST.

        Example:
        .. code-block:: python

            editor.add_instructions(
                ".STEP run -1 1023 1",
                ".dc V1 -5 5"
            )

        :param instructions: Argument list of instructions to add
        :type instructions: argument list
        :returns: Nothing
        """
        for instruction in instructions:
            self.add_instruction(instruction)
       
    @classmethod     
    def prepare_for_simulator(cls, simulator: Simulator) -> None:
        """
        Sets the library paths that should be correct for the simulator object. 
        The simulator object should have had the executable path (spice_exe) set correctly.
        
        This is especially useful in 2 cases:
        * when the simulator is running under wine, as it is difficult to detect 
        the correct library paths in that case.
        * when the editor can be used with different simulators, that have different library paths.
        
        Note:
        * you can always also set the library paths manually via `set_custom_library_paths()`
        * this method is a class method and will affect all instances of the class
        
        :param simulator: Simulator object from which the library paths will be taken.
        :type simulator: Simulator
        :returns: Nothing
        """
        if simulator is None:
            raise NotImplementedError("The prepare_for_simulator method requires a simulator object")
        cls.simulator_lib_paths = simulator.get_default_library_paths()
        return
    
    @classmethod
    def set_custom_library_paths(cls, paths: Union[str, List[str]]) -> None:
        """
        Set the given library search paths to the list of directories to search when needed.
        It will delete any previous list of custom paths, but will not affect the default paths 
        (be it from `init()` or from `prepare_for_simulator()`).
        
        Note that this method is a class method and will affect all instances of the class.

        :param paths: Path(s) to add to the Search path
        :type paths: Union[str, List[str]]
        :return: Nothing    
        """
        # empty the list
        cls.custom_lib_paths = []
        # and then fill it with the new paths
        if isinstance(paths, str):
            cls.custom_lib_paths.append(paths)
        elif isinstance(paths, list):
            cls.custom_lib_paths.extend(paths)        




# coding=utf-8
# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        base_schematic.py
# Purpose:     Base classes for schematic editors
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------


import dataclasses
import enum
from typing import List, Callable, Union, Tuple
from collections import OrderedDict
import logging
from .base_editor import BaseEditor, Component, ComponentNotFoundError, SUBCKT_DIVIDER

__author__ = "Nuno Canto Brum <nuno.brum@gmail.com>"
__copyright__ = "Copyright 2021, Fribourg Switzerland"

_logger = logging.getLogger("spicelib.BaseSchematic")


class ERotation(enum.IntEnum):
    """Component Rotation Enum"""
    R0 = 0  # 0 Rotation
    R45 = 45  # 45 Rotation
    R90 = 90  # 90 Rotation
    R135 = 135  # 135 Rotation
    R180 = 180  # 180 Rotation
    R225 = 225  # 225 Rotation
    R270 = 270  # 270 Rotation
    R315 = 315  # 315 Rotation
    M0 = 360 + 0  # Mirror 0 Rotation
    M45 = 360 + 45  # Mirror 45 Rotation
    M90 = 360 + 90  # Mirror 90 Rotation
    M135 = 360 + 135  # Mirror 135 Rotation
    M180 = 360 + 180  # Mirror 180 Rotation
    M225 = 360 + 225  # Mirror 225 Rotation
    M270 = 360 + 270  # Mirror 270 Rotation
    M315 = 360 + 315  # Mirror 315 Rotation

    def __str__(self):
        if self.value == 0:
            return "0 Rotation"
        elif self.value == 45:
            return "45 Rotation"
        elif self.value == 90:
            return "90 Rotation"
        elif self.value == 135:
            return "135 Rotation"
        elif self.value == 180:
            return "180 Rotation"
        elif self.value == 225:
            return "225 Rotation"
        elif self.value == 270:
            return "270 Rotation"
        elif self.value == 315:
            return "315 Rotation"
        elif self.value == 360 + 0:
            return "Mirror 0 Rotation"
        elif self.value == 360 + 45:
            return "Mirror 45 Rotation"
        elif self.value == 360 + 90:
            return "Mirror 90 Rotation"
        elif self.value == 360 + 135:
            return "Mirror 135 Rotation"
        elif self.value == 360 + 180:
            return "Mirror 180 Rotation"
        elif self.value == 360 + 225:
            return "Mirror 225 Rotation"
        elif self.value == 360 + 270:
            return "Mirror 270 Rotation"
        elif self.value == 360 + 315:
            return "Mirror 315 Rotation"

    # def mirror_y_axis(self):
    #     if self == ERotation.R0:
    #         return ERotation.M180
    #     elif self == ERotation.R90:
    #         return ERotation.M270
    #     elif self == ERotation.R180:
    #         return ERotation.M0
    #     elif self == ERotation.R270:
    #         return ERotation.M90
    #     elif self == ERotation.M0:
    #         return ERotation.R180
    #     elif self == ERotation.M90:
    #         return ERotation.R270
    #     elif self == ERotation.M180:
    #         return ERotation.R0
    #     elif self == ERotation.M270:
    #         return ERotation.R90
    #     else:
    #         return self
    #
    # def mirror_x_axis(self):
    #     return ERotation((((self.value + 180) % 360) + 360) % 720)

    def __add__(self, rotation: 'ERotation'):
        return (self.value + rotation.value) % 360


class HorAlign(enum.Enum):
    """Horizontal Alignment Enum"""
    LEFT = "Left"
    RIGHT = "Right"
    CENTER = "Center"


class VerAlign(enum.Enum):
    """Vertical Alignment Enum"""
    TOP = "Top"
    CENTER = "Center"
    BOTTOM = "Bottom"


class TextTypeEnum(enum.IntEnum):
    """Text Type Enum"""
    NULL = enum.auto()
    COMMENT = enum.auto()
    DIRECTIVE = enum.auto()
    LABEL = enum.auto()
    ATTRIBUTE = enum.auto()
    PIN = enum.auto()  # pin label


class LineStyle:
    """Line style : width, color and pattern (dashed, dotted, etc...) """
    def __init__(self, width: str = "", color: str = "", pattern: str = ""):
        self.width: str = width
        self.color: str = color
        self.pattern: str = pattern


class Point:
    """X, Y coordinates"""
    def __init__(self, X: float, Y: float):
        self.X = X
        self.Y = Y


class Line:
    """X1, Y1, X2, Y2 coordinates"""
    def __init__(self, v1: Point, v2: Point, style: LineStyle = None, net: str = ""):
        self.V1 = v1
        self.V2 = v2
        if style is None:
            self.style = LineStyle()
        else:
            self.style = style
        self.net = net

    def touches(self, point: Point) -> bool:
        """Returns True if the line passes through the given point"""
        if self.V1.X == self.V2.X:
            if self.V1.X == point.X:
                if min(self.V1.Y, self.V2.Y) <= point.Y <= max(self.V1.Y, self.V2.Y):
                    return True
        elif self.V1.Y == self.V2.Y:
            if self.V1.Y == point.Y:
                if min(self.V1.X, self.V2.X) <= point.X <= max(self.V1.X, self.V2.X):
                    return True
        else:
            # The time saving tricks are over, the line is oblique, so, we have to do the math
            # The line is defined by the equation y = m*x + b
            # where m is the slope and b is the y intercept
            m = (self.V2.Y - self.V1.Y) / (self.V2.X - self.V1.X)
            b = self.V1.Y - m * self.V1.X
            # Now we can calculate the Y value for the given X
            y = m * point.X + b
            # If the Y value is the same as the point Y, then the line passes through the point
            if y == point.Y:
                # Now we have to check if the point is within the line segment
                if min(self.V1.X, self.V2.X) <= point.X <= max(self.V1.X, self.V2.X):
                    return True
        return False

    def intercepts(self, line: 'Line') -> bool:
        """Returns True if the line intercepts the given line.
        The intercepts is calculated by checking if the line touches any of the line vertices
        """
        # We have to check if the line touches any of the vertices of the given line
        if self.touches(line.V1) or self.touches(line.V2):
            return True
        # We also have to check if the given line touches any of the vertices of this line
        if line.touches(self.V1) or line.touches(self.V2):
            return True
        return False


class Shape:
    """Polygon object. The shape is defined by a list of points. It can define a closed or open shape.
    The closed shape is defined by the first and last points being the same. In this case, it can have a fill.
    It is used to define polygons, arcs, circles or more complex shapes like the ones found in QSPICE"""
    def __init__(self, name: str, points: List[Point], line_style: LineStyle = None, fill: str = ""):
        self.name = name
        self.points = points
        if line_style is None:
            self.line_style = LineStyle()
        else:
            self.line_style = line_style
        self.fill = fill

# Rectangle = Shape
# Rectangle is a special case of a Shape. Rectangle is defined by two points that define the diagonal
# of the rectangle.

# Circle = Shape  # Circle is a special case of a Shape. Circle is defined by the rectangle that encloses it.

# Arc = Shape
# Arc is a special case of a Shape. Since different tools have different ways of defining arcs, we will use
# # the Shape class to define them. We will only store the list of points that are provided by the tool.

#  TODO: The following code is commented out because it is not used in the current implementation. It is kept here for
# when we decode how ARCs are stored and we could use it to update them.
# @dataclasses.dataclass
# class Arc:
#     """Opting for a non-native representation of the arc as LTspice and Qspice have different
#     ways of storing Arc information. Start and Stop points are calculated as a fraction of the radius
#     for X and Y. This avoids having to deal with the calculation of sines and cosines and their inverse
#     into radians."""
#     center: Point
#     radius: float
#     start: Point = Point(0, 0)
#     stop: Point = Point(0, 0)
#     style: LineStyle = LineStyle()
#     # The Arcs are decorative, they don't have associated nets


@dataclasses.dataclass
class Text:
    """Text object"""
    coord: Point
    text: str
    size: int = 1
    type: TextTypeEnum = TextTypeEnum.NULL
    textAlignment: HorAlign = HorAlign.LEFT
    verticalAlignment: VerAlign = VerAlign.CENTER
    angle: ERotation = ERotation.R0
    visible: bool = True


@dataclasses.dataclass
class Port:
    text: Text
    direction: str


class SchematicComponent(Component):
    """Holds component information"""

    def __init__(self, parent, line):
        super().__init__(parent, line)
        self.position: Point = Point(0, 0)
        self.rotation: ERotation = ERotation.R0
        self.symbol = None

    def __str__(self):
        return f"{self.reference} {self.position.X} {self.position.Y} {self.rotation}"


class BaseSchematic(BaseEditor):
    """
    This defines the primitives (protocol) to be used for both SpiceEditor and AscEditor
    classes.
    """

    def __init__(self):
        self.components: OrderedDict[str, SchematicComponent] = OrderedDict()
        self.wires: List[Line] = []
        self.labels: List[Text] = []
        self.directives: List[Text] = []
        self.ports: List[Port] = []
        self.lines: List[Line] = []
        self.shapes: List[Shape] = []
        self.updated = False  # indicates if an edit was done and the file has to be written back to disk

    def reset_netlist(self, create_blank: bool = False) -> None:
        """Resets the netlist to the original state"""
        self.components.clear()
        self.wires.clear()
        self.labels.clear()
        self.directives.clear()
        self.lines.clear()
        self.shapes.clear()
        self.updated = False

    def copy_from(self, editor: 'BaseSchematic') -> None:
        """Clones the contents of the given editor"""
        from copy import deepcopy
        self.components = deepcopy(editor.components)
        self.wires = deepcopy(editor.wires)
        self.labels = deepcopy(editor.labels)
        self.directives = deepcopy(editor.directives)
        self.lines = deepcopy(editor.lines)
        self.shapes = deepcopy(editor.shapes)
        self.updated = True

    def _get_parent(self, reference) -> Tuple["BaseSchematic", str]:
        if SUBCKT_DIVIDER in reference:
            sub_ref, sub_comp = reference.split(SUBCKT_DIVIDER, 1)

            subckt = self.get_component(sub_ref)
            subcircuit = subckt.attributes['_SUBCKT']
            return subcircuit, sub_comp
        else:
            return self, reference

    def set_updated(self, reference):
        sub_circuit, _ = self._get_parent(reference)
        sub_circuit.updated = True

    def get_component(self, reference: str) -> SchematicComponent:
        """
        Returns the SchematicComponent object representing the given reference in the schematic file.

        :param reference: The reference of the component
        :return: The SchematicComponent object
        :raises ComponentNotFoundError: If the component is not found.
        """
        sub_circuit, ref = self._get_parent(reference)

        if sub_circuit != self:  # The component is in a subcircuit
            return sub_circuit.get_component(ref)
        else:
            if ref not in sub_circuit.components:
                _logger.error(f"Component {reference} not found")
                raise ComponentNotFoundError(f"Component {reference} not found in Schematic file")
            return sub_circuit.components[ref]

    def get_component_position(self, reference: str) -> Tuple[Point, ERotation]:
        """Returns the position and rotation of the component"""
        comp = self.get_component(reference)
        return comp.position, comp.rotation

    def set_component_position(self, reference: str, position: Point, rotation: ERotation) -> None:
        """Sets the position and rotation of the component.

        :param reference: The reference of the component
        :type reference: str
        :param position: The new position of the component
        :type position: Point
        :param rotation: The new rotation of the component
        :type rotation: ERotation
        """
        comp = self.get_component(reference)
        comp.position = position
        comp.rotation = rotation
        self.set_updated(reference)

    def add_component(self, component: SchematicComponent, **kwargs) -> None:
        self.components[component.reference] = component
        component.parent = self
        if kwargs:
            # Update attributes
            component.attributes.update(kwargs)
        self.updated = True

    def scale(self, offset_x, offset_y, scale_x, scale_y: float,
              round_fun: Callable[[float], Union[int, float]] = None) -> None:
        """Scales the schematic"""
        if round_fun is None:
            round_fun = int
        for comp in self.components.values():
            comp.position.X = round_fun(comp.position.X * scale_x + offset_x)
            comp.position.Y = round_fun(comp.position.Y * scale_y + offset_y)
        for wire in self.wires:
            wire.V1.X = round_fun(wire.V1.X * scale_x + offset_x)
            wire.V1.Y = round_fun(wire.V1.Y * scale_y + offset_y)
            wire.V2.X = round_fun(wire.V2.X * scale_x + offset_x)
            wire.V2.Y = round_fun(wire.V2.Y * scale_y + offset_y)
        for label in self.labels:
            label.coord.X = round_fun(label.coord.X * scale_x + offset_x)
            label.coord.Y = round_fun(label.coord.Y * scale_y + offset_y)
        for directive in self.directives:
            directive.coord.X = round_fun(directive.coord.X * scale_x + offset_x)
            directive.coord.Y = round_fun(directive.coord.Y * scale_y + offset_y)
        for port in self.ports:
            port.text.coord.X = round_fun(port.text.coord.X * scale_x + offset_x)
            port.text.coord.Y = round_fun(port.text.coord.Y * scale_y + offset_y)
        for line in self.lines:
            line.V1.X = round_fun(line.V1.X * scale_x + offset_x)
            line.V1.Y = round_fun(line.V1.Y * scale_y + offset_y)
            line.V2.X = round_fun(line.V2.X * scale_x + offset_x)
            line.V2.Y = round_fun(line.V2.Y * scale_y + offset_y)
        for shape in self.shapes:
            for point in shape.points:
                point.X = round_fun(point.X * scale_x + offset_x)
                point.Y = round_fun(point.Y * scale_y + offset_y)
        self.updated = True



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        test_asc_editor.py
# Purpose:     Tool used validate the LTSpice ASC Files Editor
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------

import os
import sys
import unittest

sys.path.append(
    os.path.abspath((os.path.dirname(os.path.abspath(__file__)) + "/../")))  # add project root to lib search path

import spicelib

test_dir = '../examples/testfiles/' if os.path.abspath(os.curdir).endswith('unittests') else './examples/testfiles/'
golden_dir = './golden/' if os.path.abspath(os.curdir).endswith('unittests') else './unittests/golden/'
temp_dir = './temp/' if os.path.abspath(os.curdir).endswith('unittests') else './unittests/temp/'

if not os.path.exists(temp_dir):
    os.mkdir(temp_dir)


class ASC_Editor_Test(unittest.TestCase):

    def setUp(self):
        self.edt = spicelib.editor.asc_editor.AscEditor(test_dir + "DC sweep.asc")

    def test_component_editing(self):
        r1 = self.edt['R1']
        self.assertEqual('10k', r1.value_str, "Tested R1 Value")  # add assertion here
        self.assertListEqual(['Vin', 'R1', 'R2', 'D1'], self.edt.get_components(), "Tested get_components")  # add assertion here
        r1.value = 33000
        self.edt.save_netlist(temp_dir + 'test_components_output.asc')
        self.equalFiles(temp_dir + 'test_components_output.asc', golden_dir + 'test_components_output.asc')
        self.assertEqual(self.edt.get_component_value('R1'), '33k', "Tested R1 Value")  # add assertion here
        self.assertEqual(r1.value_str, '33k', "Tested R1 Value")  # add assertion here
        r1.set_params(Tc1='0', Tc2='0', pwr=None)
        # self.edt.set_component_parameters('R1', Tc1='0', Tc2='0', pwr=None)
        self.edt.save_netlist(temp_dir + 'test_components_output_2.asc')
        self.equalFiles(temp_dir + 'test_components_output_2.asc', golden_dir + 'test_components_output_2.asc')
        for key, value in {'Tc1': 0, 'Tc2': 0}.items():
            self.assertEqual(r1.params[key], value, f"Tested R1 {key} Parameter")
        self.edt.remove_component('R1')
        self.edt.save_netlist(temp_dir + 'test_components_output_1.asc')
        self.equalFiles(temp_dir + 'test_components_output_1.asc', golden_dir + 'test_components_output_1.asc')

    def test_component_legacy_editing(self):
        self.assertEqual(self.edt.get_component_value('R1'), '10k', "Tested R1 Value")  # add assertion here
        self.assertListEqual(self.edt.get_components(), ['Vin', 'R1', 'R2', 'D1'], "Tested get_components")  # add assertion here
        self.edt.set_component_value('R1', '33k')
        self.edt.save_netlist(temp_dir + 'test_components_output.asc')
        self.equalFiles(temp_dir + 'test_components_output.asc', golden_dir + 'test_components_output.asc')
        self.assertEqual(self.edt.get_component_value('R1'), '33k', "Tested R1 Value")  # add assertion here
        self.edt.set_component_parameters('R1', Tc1='0', Tc2='0', pwr=None)
        self.edt.save_netlist(temp_dir + 'test_components_output_2.asc')
        self.equalFiles(temp_dir + 'test_components_output_2.asc', golden_dir + 'test_components_output_2.asc')
        r1_params = self.edt.get_component_parameters('R1')
        for key, value in {'Tc1': 0, 'Tc2': 0}.items():
            self.assertEqual(r1_params[key], value, f"Tested R1 {key} Parameter")
        self.edt.remove_component('R1')
        self.edt.save_netlist(temp_dir + 'test_components_output_1.asc')
        self.equalFiles(temp_dir + 'test_components_output_1.asc', golden_dir + 'test_components_output_1.asc')


    def test_parameter_edit(self):
        self.assertEqual(self.edt.get_parameter('TEMP'), '0', "Tested TEMP Parameter")  # add assertion here
        self.edt.set_parameter('TEMP', 25)
        self.assertEqual(self.edt.get_parameter('TEMP'), '25', "Tested TEMP Parameter")  # add assertion here
        self.edt.save_netlist(temp_dir + 'test_parameter_output.asc')
        self.equalFiles(temp_dir + 'test_parameter_output.asc', golden_dir + 'test_parameter_output.asc')
        self.edt.set_parameter('TEMP', 0)  # reset to 0
        self.assertEqual(self.edt.get_parameter('TEMP'), '0', "Tested TEMP Parameter")  # add assertion here

    def test_instructions(self):
        self.edt.add_instruction('.ac dec 10 1 100k')
        self.edt.add_instruction('.save V(vout)')
        self.edt.add_instruction('.save I(R1)')
        self.edt.add_instruction('.save I(R2)')
        self.edt.add_instruction('.save I(D1)')
        self.edt.save_netlist(temp_dir + 'test_instructions_output.asc')
        self.equalFiles(temp_dir + 'test_instructions_output.asc', golden_dir + 'test_instructions_output.asc')
        self.edt.remove_instruction('.save I(R1)')
        self.edt.save_netlist(temp_dir + 'test_instructions_output_1.asc')
        self.equalFiles(temp_dir + 'test_instructions_output_1.asc', golden_dir + 'test_instructions_output_1.asc')
        self.edt.remove_Xinstruction(r"\.save\sI\(.*\)")  # removes all .save instructions for currents
        self.edt.save_netlist(temp_dir + 'test_instructions_output_2.asc')
        self.equalFiles(temp_dir + 'test_instructions_output_2.asc', golden_dir + 'test_instructions_output_2.asc')

    def equalFiles(self, file1, file2):
        with open(file1, 'r') as f1:
            lines1 = f1.readlines()
        with open(file2, 'r') as f2:
            lines2 = f2.readlines()
        self.assertEqual(len(lines1), len(lines2), "Number of lines is different")
        for i, lines in enumerate(zip(lines1, lines2)):
            self.assertEqual(lines[0], lines[1], "Line %d" % i)


if __name__ == '__main__':
    unittest.main()



from spicelib import RawRead

from matplotlib import pyplot as plt

rawfile = RawRead("./testfiles/TRAN - STEP.raw")

print(rawfile.get_trace_names())
print(rawfile.get_raw_property())

IR1 = rawfile.get_trace("I(R1)")
x = rawfile.get_trace('time')  # Gets the time axis
steps = rawfile.get_steps()
for step in range(len(steps)):
    # print(steps[step])
    plt.plot(x.get_wave(step), IR1.get_wave(step), label=steps[step])

plt.legend()  # order a legend
plt.show()



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        ltspice_utils.py
# Purpose:     Utility functions for LTSpice files
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     28-03-2024
# Licence:     refer to the LICENSE file
#
# -------------------------------------------------------------------------------

import re

from .base_schematic import ERotation, Text, HorAlign, VerAlign

__author__ = "Nuno Canto Brum <nuno.brum@gmail.com>"
__copyright__ = "Copyright 2024, Fribourg Switzerland"


# Regular expressions
TEXT_REGEX = re.compile(
    r"TEXT (-?\d+)\s+(-?\d+)\s+V?(Left|Right|Top|Bottom|Center|Invisible)\s(\d+)\s*(?P<type>[!;])(?P<text>.*)",
    re.IGNORECASE
)

TEXT_REGEX_X = 1
TEXT_REGEX_Y = 2
TEXT_REGEX_ALIGN = 3
TEXT_REGEX_SIZE = 4
TEXT_REGEX_TYPE = 5
TEXT_REGEX_TEXT = 6
END_LINE_TERM = "\n"
ASC_ROTATION_DICT = {
    'R0': ERotation.R0,
    'R90': ERotation.R90,
    'R180': ERotation.R180,
    'R270': ERotation.R270,
    'M0': ERotation.M0,
    'M90': ERotation.M90,
    'M180': ERotation.M180,
    'M270': ERotation.M270,
}
ASC_INV_ROTATION_DICT = {val: key for key, val in ASC_ROTATION_DICT.items()}
LT_ATTRIBUTE_NUMBERS = {
    'Prefix': 0,
    'Type': 1,
    'Value': 3,
    'Value2': 123,
    'SpiceModel': 38,
    'ModelFile': 'X',
    'Def_Sub': 'X',
    'SpiceLine': 39,
    'SpiceLine2': 40,
}
LT_ATTRIBUTE_NUMBERS_INV = {val: key for key, val in LT_ATTRIBUTE_NUMBERS.items()}
# not used right now, LTSpice does not seem to support it
WEIGHT_CONVERSION_TABLE = ('Thin', 'Normal', 'Thick')


def asc_text_align_set(text: Text, alignment: str):
    """Sets the alignment of the text in the ASC format"""
    # Default
    text.textAlignment = HorAlign.CENTER
    text.verticalAlignment = VerAlign.CENTER
    if alignment == 'Left' or alignment == 'VLeft':
        text.textAlignment = HorAlign.LEFT
    elif alignment == 'Center' or alignment == 'VCenter':
        pass
    elif alignment == 'Right' or alignment == 'VRight':
        text.textAlignment = HorAlign.RIGHT
    elif alignment == 'Top' or alignment == 'VTop':
        text.verticalAlignment = VerAlign.TOP
    elif alignment == 'Bottom' or alignment == 'VBottom':
        text.verticalAlignment = VerAlign.BOTTOM
    elif alignment == 'Invisible' or alignment == 'VInvisible':
        text.visible = False
    else:
        raise ValueError(f"Invalid alignment {alignment}")

    if alignment.startswith('V'):
        text.angle = ERotation.R90
    else:
        text.angle = ERotation.R0
    return text


def asc_text_align_get(text: Text) -> str:
    """Returns the alignment of the text in the ASC format"""
    if not text.visible:
        ans = 'Invisible'
    else:
        if text.verticalAlignment == VerAlign.CENTER:
            if text.textAlignment == HorAlign.RIGHT:
                ans = 'Right'
            elif text.textAlignment == HorAlign.CENTER:
                ans = 'Center'
            else:
                ans = 'Left'
        else:
            if text.verticalAlignment == VerAlign.TOP:
                ans = 'Top'
            elif text.verticalAlignment == VerAlign.BOTTOM:
                ans = 'Bottom'
            else:            
                ans = 'Left'

    if text.angle == ERotation.R90:
        ans = 'V' + ans
    return ans



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        test_qspice_rawread.py
# Purpose:     Unit testing for RawRead class using Qspice simulation resutls
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
"""
@author:        Nuno Brum
@copyright:     Copyright 2022
@credits:       nunobrum

@license:       GPLv3
@maintainer:    Nuno Brum
@email:         me@nunobrum.com

@file:          test_spicelib.py
@date:          2022-09-19

@note           spicelib ltsteps + sim_commander + raw_read unit test
                  run ./test/unittests/test_spicelib
"""

import os  # platform independent paths
# ------------------------------------------------------------------------------
# Python Libs
import sys  # python path handling
import unittest  # performs test

#
# Module libs

sys.path.append(
    os.path.abspath((os.path.dirname(os.path.abspath(__file__)) + "/../")))  # add project root to lib search path
from spicelib.log.qspice_log_reader import QspiceLogReader
from spicelib.raw.raw_read import RawRead
from spicelib.editor.spice_editor import SpiceEditor
from spicelib.sim.sim_runner import SimRunner


def has_qspice_detect():
    from spicelib.simulators.qspice_simulator import Qspice
    global qspice_simulator
    qspice_simulator = Qspice
    return qspice_simulator.is_available()


# ------------------------------------------------------------------------------
has_qspice = has_qspice_detect()
skip_qspice_editor_tests = not has_qspice
# print("skip_ltspice_tests", skip_ltspice_tests)
test_dir = '../examples/testfiles/' if os.path.abspath(os.curdir).endswith('unittests') else './examples/testfiles/'
# test_dir = os.path.abspath(test_dir)
print("test_dir", test_dir)
# ------------------------------------------------------------------------------


class test_spicelib(unittest.TestCase):
    """Unnittesting spicelib"""
    # *****************************
    @unittest.skipIf(skip_qspice_editor_tests, "Skip if not in windows environment")
    def test_batch_test(self):
        """
        @note   inits class
        """
        print("Starting test_batch_test")
        from spicelib.simulators.qspice_simulator import Qspice
        # prepare
        self.sim_files = []
        self.measures = {}

        def processing_data(raw_file, log_file):
            print("Handling the simulation data of %s, log file %s" % (raw_file, log_file))
            self.sim_files.append((raw_file, log_file))

        # select spice model
        editor = SpiceEditor(test_dir + "QSPICE_Batch_Test.net")
        runner = SimRunner(parallel_sims=4, output_folder="./output", simulator=Qspice)
        editor.set_parameters(res=0, cap=100e-6)
        editor.set_component_value('R2', '2k')  # Modifying the value of a resistor
        editor.set_component_value('R1', '4k')
        editor.set_element_model('V3', "SINE(0 1 3k 0 0 0)")  # Modifying the

        editor.set_parameter("run", "0")

        for supply_voltage in (5, 10, 15):
            editor.set_component_value('V1', supply_voltage)
            editor.set_component_value('V2', -supply_voltage)
            # overriding the automatic netlist naming
            run_netlist_file = "{}_{}.net".format(editor.circuit_file.name, supply_voltage)
            runner.run(editor, run_filename=run_netlist_file, callback=processing_data)

        runner.wait_completion()
        self.assertEqual(runner.okSim, 3)  # Simulation done with success
        # Sim Statistics
        print('Successful/Total Simulations: ' + str(runner.okSim) + '/' + str(runner.runno))
        self.assertEqual(runner.okSim, 3)
        self.assertEqual(runner.runno, 3)

        # check
        editor.reset_netlist()
        editor.set_element_model('V3', "AC 1 0")
        editor.add_instructions(
                "; Simulation settings",
                ".ac dec 30 1 10Meg",
                ".meas AC GainAC MAX mag(V(out)) ; find the peak response and call it ""Gain""",
                ".meas AC FcutAC TRIG mag(V(out))=GainAC/sqrt(2) FALL=last",
                # ".step param run 0 2 1"
        )

        raw_file, log_file = runner.run_now(editor, run_filename="qspice_no_callback.net")
        print("no_callback", raw_file, log_file)
        log = QspiceLogReader(log_file)
        for measure in log.get_measure_names():
            print(measure, '=', log.get_measure_value(measure))
        self.assertEqual(6.16683e+06, log.get_measure_value('fcutac')[0])
        self.assertEqual(1.99999, log.get_measure_value('gainac')[0])

    @unittest.skipIf(skip_qspice_editor_tests, "Skip if not in windows environment")
    def test_run_from_spice_editor(self):
        """Run command on SpiceEditor"""
        print("Starting test_run_from_spice_editor")
        runner = SimRunner(output_folder='temp' + "temp/", simulator=qspice_simulator)
        # select spice model
        netlist = SpiceEditor(test_dir + "testfile.net")
        # set default arguments
        netlist.set_parameters(res=0.001, cap=100e-6)
        # define simulation
        netlist.add_instructions(
                "; Simulation settings",
                # [".STEP PARAM Rmotor LIST 21 28"],
                ".TRAN 3m",
                # ".step param run 1 2 1"
        )
        # do parameter sweep
        for res in range(5):
            # LTC.runs_to_do = range(2)
            netlist.set_parameters(ANA=res)
            raw, log = runner.run(netlist).wait_results()
            print("Raw file '%s' | Log File '%s'" % (raw, log))
        # Sim Statistics
        runner.wait_completion()
        print('Successful/Total Simulations: ' + str(runner.okSim) + '/' + str(runner.runno))
        self.assertEqual(runner.okSim, 5)  # Simulation done with success

    @unittest.skipIf(skip_qspice_editor_tests, "Skip if not in windows environment")
    def test_sim_runner(self):
        """SimRunner and SpiceEditor singletons"""
        print("Starting test_sim_runner")
        # Old legacy class that merged SpiceEditor and SimRunner
        def callback_function(raw_file, log_file):
            print("Handling the simulation data of %s, log file %s" % (raw_file, log_file))

        runner = SimRunner(output_folder='temp' + "temp/", simulator=qspice_simulator)
        SE = SpiceEditor(test_dir + "testfile.net")
        #, parallel_sims=1)
        tstart = 0
        for tstop in (2, 5, 8, 10):
            tduration = tstop - tstart
            SE.add_instruction(".tran {}".format(tduration), )
            if tstart != 0:
                SE.add_instruction(".loadbias {}".format(bias_file))
                # Put here your parameter modifications
                # runner.set_parameters(param1=1, param2=2, param3=3)
            bias_file = test_dir + "sim_loadbias_%d.txt" % tstop
            SE.add_instruction(".savebias {} internal time={}".format(bias_file, tduration))
            tstart = tstop
            runner.run(SE, callback=callback_function)

        SE.reset_netlist()
        SE.add_instruction('.ac dec 40 1m 1G')
        SE.set_component_value('V1', 'AC 1 0')
        runner.run(SE, callback=callback_function)
        runner.wait_completion()
        self.assertEqual(runner.okSim, 5)  # Simulation done with success

    @unittest.skipIf(skip_qspice_editor_tests, "Execute All")
    def test_ltsteps_measures(self):
        """QspiceLogReader Measures from Batch_Test.asc"""
        print("Starting test_ltsteps_measures")
        assert_data = {
            'vout1m'   : [
                -0.0186257,
                -1.04378,
                -1.64283,
                -0.622014,
                1.32386,
                -1.35125,
                -1.88222,
                1.28677,
                1.03154,
                0.953548,
                -0.192821,
                -1.42535,
                0.451607,
                0.0980979,
                1.55525,
                1.66809,
                0.11246,
                0.424023,
                -1.30035,
                0.614292,
                -0.878185,
            ],
            'vin_rms'  : [
                0.706221,
                0.704738,
                0.708225,
                0.707042,
                0.704691,
                0.704335,
                0.704881,
                0.703097,
                0.70322,
                0.703915,
                0.703637,
                0.703558,
                0.703011,
                0.702924,
                0.702944,
                0.704121,
                0.704544,
                0.704193,
                0.704236,
                0.703701,
                0.703436,
            ],
            'vout_rms' : [
                1.41109,
                1.40729,
                1.41292,
                1.40893,
                1.40159,
                1.39763,
                1.39435,
                1.38746,
                1.38807,
                1.38933,
                1.38759,
                1.38376,
                1.37771,
                1.37079,
                1.35798,
                1.33252,
                1.24314,
                1.07237,
                0.875919,
                0.703003,
                0.557131,

            ],
            'gain'     : [
                1.99809,
                1.99689,
                1.99502,
                1.99271,
                1.98894,
                1.98432,
                1.97814,
                1.97336,
                1.97387,
                1.97372,
                1.97202,
                1.9668,
                1.95973,
                1.95012,
                1.93184,
                1.89246,
                1.76445,
                1.52284,
                1.24379,
                0.999007,
                0.792014,
            ],
            'period'   : [
                0.000100148,
                7.95811e-005,
                6.32441e-005,
                5.02673e-005,
                3.99594e-005,
                3.1772e-005,
                2.52675e-005,
                2.01009e-005,
                1.59975e-005,
                1.27418e-005,
                1.01541e-005,
                8.10036e-006,
                6.47112e-006,
                5.18241e-006,
                4.16639e-006,
                3.37003e-006,
                2.75114e-006,
                2.26233e-006,
                1.85367e-006,
                1.50318e-006,
                1.20858e-006,

            ],
            'period_at': [
                0.000100148,
                7.95811e-005,
                6.32441e-005,
                5.02673e-005,
                3.99594e-005,
                3.1772e-005,
                2.52675e-005,
                2.01009e-005,
                1.59975e-005,
                1.27418e-005,
                1.01541e-005,
                8.10036e-006,
                6.47112e-006,
                5.18241e-006,
                4.16639e-006,
                3.37003e-006,
                2.75114e-006,
                2.26233e-006,
                1.85367e-006,
                1.50318e-006,
                1.20858e-006,
            ]
        }
        if has_qspice:
            runner = SimRunner(output_folder='temp', simulator=qspice_simulator)
            raw_file, log_file = runner.run_now(test_dir + "QSPICE_Batch_Test.net")
            print(raw_file, log_file)
        else:
            log_file = test_dir + "QSPICE_Batch_Test_1.log"
        log = QspiceLogReader(log_file)
        # raw = RawRead(raw_file)
        for measure in assert_data:
            print("measure", measure)
            for step in range(log.step_count):
                self.assertEqual(log.get_measure_value(measure, step), assert_data[measure][step])

                print(log.get_measure_value(measure, step), assert_data[measure][step])

    @unittest.skipIf(False, "Execute All")
    def test_operating_point(self):
        """Operating Point Simulation Test"""
        print("Starting test_operating_point")
        if has_qspice:
            runner = SimRunner(output_folder='temp', simulator=qspice_simulator)
            raw_file, log_file = runner.run_now(test_dir + "DC op point.net")
            self.assertEqual(runner.okSim, 1)  # Simulation done with success
        else:
            raw_file = test_dir + "DC op point_1.qraw"
            # log_file = test_dir + "DC op point_1.log"
        raw = RawRead(raw_file)

        for trace, value_expected  in zip(
                ('V(in)', 'V(out)', 'I(R1)', 'I(R2)', 'I(Vin)'),
                (1.0, 0.5, 5e-05, 5e-05, 5e-05)):
            value_read = raw.get_trace(trace)[0]
            self.assertAlmostEqual(abs(value_read), value_expected, 5, f"mismatch in node {trace}")

    @unittest.skipIf(False, "Execute All")
    def test_operating_point_step(self):
        """Operating Point Simulation with Steps """
        print("Starting test_operating_point_step")
        if has_qspice:
            runner = SimRunner(output_folder='temp', simulator=qspice_simulator)
            raw_file, log_file = runner.run_now(test_dir + "DC op point - STEP.net")
        else:
            raw_file = test_dir + "DC op point - STEP_1.qraw"
        raw = RawRead(raw_file)
        vin = raw.get_trace('V(in)')

        for i, b in enumerate(('V(in)', 'V(b4)', 'V(b3)', 'V(b2)', 'V(b1)', 'V(b0)'),):
            meas = raw.get_trace(b)
            for step in range(raw.nPoints):
                self.assertAlmostEqual(meas[step], vin[step] * 2**-i, 7, f"mismatch in node {b}")

    @unittest.skipIf(skip_qspice_editor_tests, "Execute All")
    def test_transient(self):
        """Transient Simulation test """
        print("Starting test_transient")
        if has_qspice:
            runner = SimRunner(output_folder='temp', simulator=qspice_simulator)
            raw_file, log_file = runner.run_now(test_dir + "TRAN.net")
        else:
            raw_file = test_dir + "TRAN_1.raw"
            log_file = test_dir + "TRAN_1.log"
        raw = RawRead(raw_file)
        log = QspiceLogReader(log_file)
        vout = raw.get_trace('V(out)')
        meas = ('t1', 't2', 't3', 't4', 't5',)
        time = (1e-3, 2e-3, 3e-3, 4e-3, 5e-3,)
        for m, t in zip(meas, time):
            log_value = log.get_measure_value(m)
            raw_value = vout.get_point_at(t)
            print(log_value, raw_value, log_value - raw_value)
            self.assertAlmostEqual(log_value, raw_value, 2, "Mismatch between log file and raw file")

    @unittest.skipIf(skip_qspice_editor_tests, "Execute All")
    def test_transient_steps(self):
        """Transient simulation with stepped data."""
        print("Starting test_transient_steps")
        if has_qspice:
            runner = SimRunner(output_folder='temp', simulator=qspice_simulator)
            raw_file, log_file = runner.run_now(test_dir + "QSPICE_TRAN - STEP.net")
        else:
            raw_file = test_dir + "QSPICE_TRAN - STEP_1.qraw"
            log_file = test_dir + "QSPICE_TRAN - STEP_1.log"

        raw = RawRead(raw_file)
        log = QspiceLogReader(log_file)
        vout = raw.get_trace('V(out)')
        meas = ('t1', 't2', 't3', 't4', 't5',)
        time = (1e-3, 2e-3, 3e-3, 4e-3, 5e-3,)
        for m, t in zip(meas, time):
            for step, step_dict in enumerate(raw.steps):
                log_value = log.get_measure_value(m, step)
                raw_value = vout.get_point_at(t, step)
                self.assertAlmostEqual(log_value, raw_value, 5, f"Mismatch between log file and raw file in step :{step_dict} measure: {m} ")

    @unittest.skipIf(False, "Execute All")
    def test_ac_analysis(self):
        """AC Analysis Test"""
        print("Starting test_ac_analysis")
        from numpy import pi, angle
        if has_qspice:
            editor = SpiceEditor(test_dir + "AC.net")
            runner = SimRunner(output_folder='temp', simulator=qspice_simulator)
            raw_file, log_file = runner.run_now(editor)

            R1 = editor.get_component_floatvalue('R1')
            C1 = editor.get_component_floatvalue('C1')
        else:
            raw_file = test_dir + "AC_1.raw"
            log_file = test_dir + "AC_1.log"
            R1 = 100
            C1 = 10E-6
        # Compute the RC AC response with the resistor and capacitor values from the netlist.
        raw = RawRead(raw_file)
        vout_trace = raw.get_trace('V(out)')
        vin_trace = raw.get_trace('V(in)')
        for point, freq in enumerate(raw.axis):
            vout1 = vout_trace.get_point_at(freq)
            vout2 = vout_trace.get_point(point)
            vin = vin_trace.get_point(point)
            self.assertEqual(vout1, vout2)
            self.assertEqual(abs(vin), 1)
            # Calculate the magnitude of the answer Vout = Vin/(1+jwRC)
            h = vin/(1 + 2j * pi * freq * R1 * C1)
            self.assertAlmostEqual(abs(vout1), abs(h), 5, f"Difference between theoretical value ans simulation at point {point}")
            self.assertAlmostEqual(angle(vout1), angle(h), 5, f"Difference between theoretical value ans simulation at point {point}")

    @unittest.skipIf(False, "Execute All")
    def test_ac_analysis_steps(self):
        """AC Analysis Test with steps"""
        print("Starting test_ac_analysis_steps")
        from numpy import pi, angle
        if has_qspice:
            editor = SpiceEditor(test_dir + "AC - STEP.net")
            runner = SimRunner(output_folder='temp', simulator=qspice_simulator)
            raw_file, log_file = runner.run_now(editor)
            C1 = editor.get_component_floatvalue('C1')
        else:
            raw_file = test_dir + "AC - STEP_1.raw"
            log_file = test_dir + "AC - STEP_1.log"
            C1 = 159.1549e-6  # 159.1549uF
        # Compute the RC AC response with the resistor and capacitor values from the netlist.
        raw = RawRead(raw_file)
        vin_trace = raw.get_trace('V(in)')
        vout_trace = raw.get_trace('V(out)')
        for step, step_dict in enumerate(raw.steps):
            R1 = step_dict['r1']
            # print(step, step_dict)
            for point in range(0, raw.get_len(step), 10):  # 10 times less points
                print(point, end=' - ')
                vout = vout_trace.get_point(point, step)
                vin = vin_trace.get_point(point, step)
                freq = raw.axis.get_point(point, step)
                # Calculate the magnitude of the answer Vout = Vin/(1+jwRC)
                h = vin/(1 + 2j * pi * freq * R1 * C1)
                # print(freq, vout, h, vout - h)
                self.assertAlmostEqual(abs(vout), abs(h), 5,
                                       f"Difference between theoretical value ans simulation at point {point}:")
                self.assertAlmostEqual(angle(vout), angle(h), 5,
                                       f"Difference between theoretical value ans simulation at point {point}")

    # 
    # def test_pathlib(self):
    #     """pathlib support"""
    #     import pathlib
    #     DIR = pathlib.Path("../tests")
    #     raw_file = DIR / "AC - STEP_1.raw"
    #     raw = RawRead(raw_file)


# ------------------------------------------------------------------------------
if __name__ == '__main__':
    print("Starting tests on spicelib")
    unittest.main()
    print("Tests completed on spicelib")
# ------------------------------------------------------------------------------



#!/usr/bin/env python
# coding=utf-8
# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        quick_sensitivity_analysis.py
# Purpose:     Classes to make a sensitivity analysis
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     16-10-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
import logging
from typing import Union, Optional, Type, Callable

from .tolerance_deviations import ToleranceDeviations, DeviationType
from ..sim_runner import AnyRunner, ProcessCallback
from ...editor.base_editor import BaseEditor
from ...log.logfile_data import LogfileData

_logger = logging.getLogger("spicelib.SimAnalysis")


class QuickSensitivityAnalysis(ToleranceDeviations):
    """Class to automate Sensitivity simulations"""
    def __init__(self, circuit_file: Union[str, BaseEditor], runner: Optional[AnyRunner] = None):
        super().__init__(circuit_file, runner)

    def prepare_testbench(self, **kwargs):
        """Prepares the simulation by setting the tolerances for each component"""
        no = 0
        self.elements_analysed.clear()
        use_min = kwargs.get('use_min', False)
        for comp in self.get_components('*'):
            val, dev = self.get_component_value_deviation_type(comp)
            new_val = val
            if dev.typ == DeviationType.tolerance:
                new_val = "{satol(%s,%g,%d)}" % (val, dev.max_val, no)
            elif dev.typ == DeviationType.minmax:
                used_value = dev.min_val if use_min else dev.max_val
                new_val = "{sammx(%s,%g,%d)}" % (val, used_value, no)

            if new_val != val:
                self.set_component_value(comp, new_val)
                self.elements_analysed.append(comp)
                no += 1

        self.last_run_number = no - 1
        if use_min:
            self.editor.add_instruction(".func satol(nom,tol,idx) nom*if(run==idx,1-tol,1)")
        else:
            self.editor.add_instruction(".func satol(nom,tol,idx) nom*if(run==idx,1+tol,1)")
        self.editor.add_instruction(".func sammx(nom,val,idx) if(run==idx,val,nom)")
        self.editor.add_instruction(".step param run -1 %d 1" % self.last_run_number)
        self.editor.set_parameter('run', -1)  # in case the step is commented.
        self.testbench_prepared = True

    def get_sensitivity_data(self, ref: str, measure: str) -> Union[float, dict, None]:
        """
        Returns the sensitivity data for a given component and measurement in terms of percentage of the total error.
        This quick approach is not very accurate, but it is fast. It assumes that the system is linear and that the
        maximum error is the sum of the absolute error of each component. This is a rough approximation, but it is
        good enough for a quick analysis. For more accurate results, use the Worst Case Analysis, which requires
        more simulation runs but gives a more accurate result.
        The best compromise, is to start with the quick analysis and then use the Worst Case Analysis to refine the
        results with only the components that have a significant contribution to the error.
        :param ref: The reference component, or '*' to return a dictionary with all the components
        :param measure: The measurement to be analysed
        :return: The sensitivity data in percentage of the total error for the reference component
        """
        if self.testbench_prepared and self.testbench_executed or self.analysis_executed:
            log_data: LogfileData = self.read_logfiles()
            nominal_data = log_data.get_measure_value(measure, run=-1)
            error_data = []
            for idx in range(len(self.elements_analysed)):
                step_data = log_data.get_measure_value(measure, run=idx)
                error_data.append(abs(step_data - nominal_data))
            total_error = sum(error_data)
            if ref == '*':
                return {ref: error_data[idx] / total_error * 100 if total_error != 0 else 0
                        for idx, ref in enumerate(self.elements_analysed)}
            else:
                idx = self.elements_analysed.index(ref)
                return error_data[idx] / total_error * 100 if total_error != 0 else 0
        else:
            _logger.warning("The analysis was not executed. Please run the run_analysis(...) or run_testbench(...)"
                            " before calling this method")
            return None

    def run_analysis(self,
                     callback: Union[Type[ProcessCallback], Callable] = None,
                     callback_args: Union[tuple, dict] = None,
                     switches=None,
                     timeout: float = None,
                     ):
        self.clear_simulation_data()
        self.elements_analysed.clear()
        # Calculate the number of runs

        worst_case_elements = {}

        def check_and_add_component(ref1: str):
            val1, dev1 = self.get_component_value_deviation_type(ref1)  # get there present value
            if dev1.min_val == dev1.max_val or dev1.typ == DeviationType.none:
                return
            worst_case_elements[ref1] = val1, dev1, 'component'
            self.elements_analysed.append(ref1)

        for ref in self.device_deviations:
            check_and_add_component(ref)

        for ref in self.parameter_deviations:
            val, dev = self.get_parameter_value_deviation_type(ref)
            if dev.typ == DeviationType.tolerance or dev.typ == DeviationType.minmax:
                worst_case_elements[ref] = val, dev, 'parameter'
                self.elements_analysed.append(ref)

        for prefix in self.default_tolerance:
            for ref in self.get_components(prefix):
                if ref not in self.device_deviations:
                    check_and_add_component(ref)

        last_run_number = len(self.elements_analysed)
        if last_run_number > 4096:
            _logger.warning("The number of runs is too high. It will be limited to 4096\n"
                            "Consider limiting the number of components with deviation")
            return

        self._reset_netlist()  # reset the netlist
        self.play_instructions()  # play the instructions
        # Add the run number to the measurements by using a parameter
        self.editor.set_parameter('run', -1)  # in case the step is commented.
        self.editor.add_instruction(".meas run PARAM {run}")
        # Run the simulation in the nominal case
        self.run(wait_resource=True,
                 callback=callback, callback_args=callback_args,
                 switches=switches, timeout=timeout)
        last_bit_setting = 0
        for run in range(last_run_number):
            # Preparing the variation on components, but only on the ones that have changed
            bit_setting = (2 ** run)
            bit_updated = bit_setting ^ last_bit_setting
            bit_index = 0
            print("bit updated: %d" % bit_updated)
            while bit_updated != 0:
                if bit_updated & 1:
                    ref = self.elements_analysed[bit_index]
                    val, dev, typ = worst_case_elements[ref]
                    if dev.typ == DeviationType.tolerance:
                        new_val = val * (1 + dev.max_val) if bit_setting & (1 << bit_index) else val
                    elif dev.typ == DeviationType.minmax:
                        new_val = dev.max_val if bit_setting & (1 << bit_index) else val
                    else:
                        _logger.warning("Unknown deviation type")
                        new_val = val
                    if typ == 'component':
                        self.editor.set_component_value(ref, new_val)  # update the value
                    elif typ == 'parameter':
                        self.editor.set_parameter(ref, new_val)
                    else:
                        _logger.warning("Unknown type")
                    print(f"{ref} = {new_val}")
                bit_updated >>= 1
                bit_index += 1
            self.editor.set_parameter('run', run)
            # Run the simulation
            self.run(wait_resource=True,
                     callback=callback, callback_args=callback_args,
                     switches=switches, timeout=timeout)
            last_bit_setting = bit_setting
        self.runner.wait_completion()

        if callback is not None:
            callback_rets = []
            for rt in self.simulations:
                callback_rets.append(rt.get_results())
            self.simulation_results['callback_returns'] = callback_rets
        self.analysis_executed = True
        # Force already the reading of logfiles
        log_data: LogfileData = self.read_logfiles()
        # if applicable, the run parameter shall be transformed into an int
        runs = []
        for run in log_data.dataset['run']:
            if isinstance(run, complex):
                runs.append(int(round(run.real, 0)))
            else:
                runs.append(run)
        log_data.dataset['run'] = runs


#!/usr/bin/env python
# coding=utf-8
import logging

import spicelib
from spicelib import AscEditor, SimRunner  # Imports the class that manipulates the asc file
from spicelib.sim.tookit.fast_worst_case import FastWorstCaseAnalysis
from spicelib.simulators.ltspice_simulator import LTspice

spicelib.set_log_level(logging.INFO)

sallenkey = AscEditor("./testfiles/sallenkey.asc")  # Reads the asc file into memory
runner = SimRunner(simulator=LTspice, output_folder='./temp_fwca', verbose=True)  # Instantiates the runner with a temp folder set
wca = FastWorstCaseAnalysis(sallenkey, runner)  # Instantiates the Worst Case Analysis class

# The following lines set the default tolerances for the components
wca.set_tolerance('R', 0.01)  # 1% tolerance
wca.set_tolerance('C', 0.1)  # 10% tolerance
# wca.set_tolerance('V', 0.1)  # 10% tolerance. For Worst Case analysis, the distribution is irrelevant
wca.set_tolerance('I', 0.1)  # 10% tolerance. For Worst Case analysis, the distribution is irrelevant
# Some components can have a different tolerance
wca.set_tolerance('R1', 0.05)  # 5% tolerance for R1 only. This only overrides the default tolerance for R1
wca.set_tolerance('R4', 0.0)  # 5% tolerance for R1 only. This only overrides the default tolerance for R1

# Tolerances can be set for parameters as well.
# wca.set_parameter_deviation('Vos', 3e-4, 5e-3)

# Finally the netlist is saved to a file
wca.save_netlist('./testfiles/sallenkey_fwc.asc')

print("=====================================")
# Now using the second method, where the simulations are ran one by one
wca.clear_simulation_data()  # Clears the simulation data
wca.reset_netlist()  # Resets the netlist to the original
nominal, min_value, max_comp_values, max_value, min_comp_values = wca.run_analysis(measure='fcut')  # Makes the Worst Case Analysis
print("Nominal Value", nominal)
print("Worst Case Min", min_value)
print("Min Component Values")
for comp, value in min_comp_values.items():
    print(f"    {comp} = {value}")
print("Worst Case Max", max_value)
print("Max Component Values")
for comp, value in max_comp_values.items():
    print(f"    {comp} = {value}")
wca.cleanup_files()  # Deletes the temporary files



#!/usr/bin/env python
# coding=utf-8
# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        montecarlo.py
# Purpose:     Classes to automate Monte-Carlo simulations
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     10-08-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------

import logging
import random
from typing import Union, Callable, Type

_logger = logging.getLogger("spicelib.SimAnalysis")

from .tolerance_deviations import ToleranceDeviations, DeviationType, ComponentDeviation
from ..process_callback import ProcessCallback
from ...log.logfile_data import LogfileData


class Montecarlo(ToleranceDeviations):
    """
    Class to automate Montecarlo simulations, where component values and parameters are replaced by a random
    distribution (either a gaussian or a uniform distribution).

    This is a statistical method, hence, it needs a considerable number of simulations to achieve a final
    gaussian distribution in order to apply this method. If the number of parameters and component values is
    low, it is better to use a Worst-Case approach.

    Like the Worst-Case and Sensitivity analysis, there are two possible approaches to use this class:

    Class to automate Worst-Case simulations, where all possible combinations of maximum and minimums
    possible values of component values and parameters are done.

    It is advised to use this algorithm when the number of parameters to be varied is reduced.
    Typically less than 10 or 12. A higher number will translate into a huge number of simulations.
    For more than 1000 simulations, it is better to use a statistical method such as the Montecarlo.

    Like the Worst-Case and Sensitivity analysis, there are two possible approaches to use this class:

        1. Preparing a testbench where all combinations are managed directly by the simulator, replacing
         parameters and component values by formulas and using a .STEP primitive to cycle through all possible
         combinations.

        2. Launching each simulation separately where the running python script manages all parameter value
        variations.

    The first approach is normally faster, but not possible in all simulators. The second approach is a valid backup
    when every single simulation takes too long, or when it is prone to crashes and stalls.
    """

    def prepare_testbench(self, **kwargs):
        """
        Prepares the simulation by setting the tolerances for the components
        :keyword num_runs: Number of runs to be performed. Default is 1000.
        :return: Nothing
        """
        min_max_uni_func = False
        min_max_norm_func = False
        tol_uni_func = False
        tol_norm_func = False
        self.elements_analysed.clear()
        for ref in self.get_components('*'):
            val, dev = self.get_component_value_deviation_type(ref)  # get there present value
            new_val = val
            if dev.typ == DeviationType.tolerance:
                tolstr = ('%g' % dev.max_val).rstrip('0').rstrip('.')
                if dev.distribution == 'uniform':
                    new_val = "{utol(%s,%s)}" % (val, tolstr)  # calculate expression for new value
                    tol_uni_func = True
                elif dev.distribution == 'normal':
                    new_val = "{ntol(%s,%s)}" % (val, tolstr)
                    tol_norm_func = True
            elif dev.typ == DeviationType.minmax:
                if dev.distribution == 'uniform':
                    new_val = "{urng(%s, %s,%s)}" % (val, dev.min_val, dev.max_val)  # calculate expression for new value
                    min_max_uni_func = True
                elif dev.distribution == 'normal':
                    new_val = "{nrng(%s,%s,%s)}" % (val, dev.min_val, dev.max_val)
                    min_max_norm_func = True

            if new_val != val:  # Only update the value if it has changed
                self.set_component_value(ref, new_val)  # update the value
                self.elements_analysed.append(ref)

        for param in self.parameter_deviations:
            val, dev = self.get_parameter_value_deviation_type(param)
            new_val = val
            if dev.typ == DeviationType.tolerance:
                if dev.distribution == 'uniform':
                    new_val = "{utol(%s,%g)}" % (val, dev.max_val)
                    tol_uni_func = True
                elif dev.distribution == 'normal':
                    new_val = "{ntol(%g,%g)}" % (val, dev.max_val)
                    tol_norm_func = True
            elif dev.typ == DeviationType.minmax:
                if dev.distribution == 'uniform':
                    new_val = "{urng(%s,%g,%g)}" % (val, (dev.max_val+dev.min_val)/2, (dev.max_val-dev.min_val)/2)
                    min_max_uni_func = True
                elif dev.distribution == 'normal':
                    new_val = "{nrng(%s,%g,%g)}" % (val, (dev.max_val+dev.min_val)/2, (dev.max_val-dev.min_val)/6)
                    min_max_norm_func = True
            else:
                continue
            self.editor.set_parameter(param, new_val)
            self.elements_analysed.append(param)

        if self.runner.simulator.__name__ == 'LTspice':
            if tol_uni_func:
                self.editor.add_instruction(".func utol(nom,tol) if(run<0, nom, mc(nom,tol))")

            if tol_norm_func:
                self.editor.add_instruction(".func ntol(nom,tol) if(run<0, nom, nom*(1+gauss(tol/3)))")

            if min_max_uni_func:
                self.editor.add_instruction(".func urng(nom,mean,df2) if(run<0, nom, mean+flat(df2))")

            if min_max_norm_func:
                self.editor.add_instruction(".func nrng(nom,mean,df6) if(run<0, nom, mean*(1+gauss(df6)))")
        elif self.runner.simulator.__name__ == 'Qspice':
            # if gauss function is needed
            #  => This is finally not needed because Qspice has a built-in gauss function (non-documented)
            # if tol_norm_func or min_max_norm_func:
            #   self.editor.add_instruction(".func random_not0() {(random()+1e-7)/(1+1e-7)}")
            #   self.editor.add_instruction(".func gauss(sigma) {sqrt(-2*ln(random_not0()))*cos(2*pi*random())*sigma}")
            if tol_uni_func:
                self.editor.add_instruction(".func utol(nom,tol) {if(run<0, nom, nom*(1+tol*(2*random()-1)))}")

            if tol_norm_func:
                self.editor.add_instruction(".func ntol(nom,tol) {if(run<0, nom, nom*(1+gauss(tol/3)))}")

            if min_max_uni_func:
                self.editor.add_instruction(".func urng(nom,mean,df2) if(run<0, nom, mean+(df2*(2*random()-1))")

            if min_max_norm_func:
                self.editor.add_instruction(".func nrng(nom,mean,df6) if(run<0, nom, mean*(1+gauss(df6)))")
        else:
            _logger.warning("Simulator not supported for this method")
            raise NotImplementedError("Simulator not supported for this method")

        self.last_run_number = kwargs.get('num_runs', self.last_run_number if self.last_run_number != 0 else 1000)
        self.editor.add_instruction(".step param run -1 %d 1" % self.last_run_number)
        self.editor.set_parameter('run', -1)
        self.testbench_prepared = True

    @staticmethod
    def _get_sim_value(value: float, dev: ComponentDeviation) -> float:
        """Returns a new value for the simulation"""
        new_val = value
        if dev.typ == DeviationType.tolerance:
            if dev.distribution == 'uniform':
                new_val = f"{random.Random().uniform(value * (1- dev.max_val), value * (1 + dev.max_val)):g}"
            elif dev.distribution == 'normal':
                new_val = f"{random.Random().gauss(value, dev.max_val / 3):g}"
        elif dev.typ == DeviationType.minmax:
            if dev.distribution == 'uniform':
                new_val = f"{random.Random().uniform(dev.min_val, dev.max_val):g}"
            elif dev.distribution == 'normal':
                new_val = f"{random.Random().gauss((dev.max_val + dev.min_val) / 2, (dev.max_val - dev.min_val) / 6):g}"
        elif dev.typ == DeviationType.none:
            pass
        else:
            _logger.warning("Unknown deviation type")
        return new_val

    def run_analysis(self,
                     callback: Union[Type[ProcessCallback], Callable] = None,
                     callback_args: Union[tuple, dict] = None,
                     switches=None,
                     timeout: float = None,
                     num_runs: int = 1000,
                     ):
        """This method runs the analysis without updating the netlist.
        It will update component values and parameters according to their deviation type and call the simulation.
        The advantage of this method is that it doesn't require adding random functions to the netlist.
        The number of times the simulation is done is specified on the argument num_runs."""
        self.elements_analysed.clear()
        self.clear_simulation_data()
        for run in range(num_runs):
            self._reset_netlist()  # reset the netlist
            self.play_instructions()  # play the instructions
            # Preparing the variation on components
            for ref in self.get_components('*'):
                val, dev = self.get_component_value_deviation_type(ref)  # get there present value
                new_val = self._get_sim_value(val, dev)
                if new_val != val:  # Only update the value if it has changed
                    self.editor.set_component_value(ref, new_val)  # update the value
            # Preparing the variation on parameters
            for param in self.parameter_deviations:
                val, dev = self.get_parameter_value_deviation_type(param)
                new_val = self._get_sim_value(val, dev)
                if new_val != val:  # Only update the value if it has changed
                    self.editor.set_parameter(param, new_val)
            # Run the simulation
            rt = self.run(wait_resource=True,
                          callback=callback, callback_args=callback_args,
                          switches=switches, timeout=timeout)

        self.runner.wait_completion()
        if callback is not None:
            callback_rets = []
            for rt in self.simulations:
                callback_rets.append(rt.get_results())
            self.simulation_results['callback_returns'] = callback_rets
        self.analysis_executed = True

    def analyse_measurement(self, meas_name: str):
        """Returns the measurement data for the given measurement name.
        If the measurement is not found, it returns None
        Note: It is up to the user to make the statistics on the data. The traditional way is to use the numpy package
        to calculate the mean and standard deviation of the data. It is also usual to consider max and min as 3 sigma,
        which is 99.7% of the data.
        """
        if not self.analysis_executed:
            _logger.warning("The analysis was not executed. Please run the analysis before calling this method")
            return None
        log_data: LogfileData = self.read_logfiles()
        meas_data = log_data[meas_name]
        if meas_data is None:
            _logger.warning("Measurement %s not found in log files", meas_name)
            return None
        else:
            return meas_data
